__all__ = (
    'ExceptionWriterContextManager', 'get_default_trace_writer_highlighter', 'render_frames_into_async',
    'render_exception_into_async', 'set_default_trace_writer_highlighter', 'set_trace_writer_highlighter',
    'write_exception_async', 'write_exception_maybe_async', 'write_exception_sync'
)

import sys, threading
from os import get_terminal_size
from sys import _getframe as get_frame, platform as PLATFORM
from threading import current_thread, get_ident

from ...utils import (
    DEFAULT_ANSI_HIGHLIGHTER, HIGHLIGHT_TOKEN_TYPES, HighlightFormatterContext, alchemy_incendiary, call, export,
    include, render_exception_into, render_frames_into
)
from ...utils.trace import _add_typed_part_into

from .event_loop import get_event_loop


EventThread = include('EventThread')


async def render_frames_into_async(frames, extend=None, *, filter=None, highlighter=None, loop=None):
    """
    Renders the given frames into a list of strings.
    
    This function provides asynchronous way of executing of ``render_frames_into``.
    
    ``render_frames_into`` contains blocking io operations, like file-reads.
    
    This method is a coroutine.
    
    Parameters
    ----------
    frames : `list` of (`FrameType`, `TraceBack`, ``FrameProxyType``)
        The frames to render.
    
    extend : `None`, `list` of `str` = `None`, Optional
        Whether the frames should be rendered into an already existing list.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    loop : `None`, ``EventThread`` = `None`, Optional (Keyword only)
        The event loop to execute the operation on. If not given, is auto-detected.
    
    Returns
    -------
    extend : `list` of `str`
        The rendered frames as a `list` of it's string parts.
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    """
    if loop is None:
        loop = get_event_loop()
    
    return await loop.run_in_executor(
        alchemy_incendiary(
            render_frames_into,
            (frames, extend),
            {'filter': filter, 'highlighter': highlighter}
        )
    )


async def render_exception_into_async(exception, extend=None, *, filter=None, highlighter=None, loop=None):
    """
    Renders the given exception's frames into a list of strings.
    
    This function provides asynchronous way of executing of ``render_exception_into``.
    
    ``render_exception_into`` calls ``render_frames_into`` which contains blocking io operations, like file-reads.
    
    This method is a coroutine.
    
    Parameters
    ----------
    exception : `BaseException`
        The exception to render.
    
    extend : `None`, `list` of `str` = `None`, Optional
        Whether the frames should be rendered into an already existing list.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    Returns
    -------
    future : ``Future`` -> `list` of `str`
        The rendered frames as a `list` of it's string parts.
    """
    if loop is None:
        loop = get_event_loop()
    
    return await loop.run_in_executor(
        alchemy_incendiary(
            render_exception_into,
            (exception, extend),
            {'filter': filter, 'highlighter': highlighter}
        )
    )


def _build_additional_title(title):
    """
    Builds additional title represented by `before` and `after` parameters of ``write_exception_sync``.
    
    Parameters
    ----------
    title : `str`, `list` of `str`
        The title to put before or after an exception traceback.
    
    Returns
    -------
    built_title : `None`, `str`
    """
    if title is None:
        return None
    
    if isinstance(title, str):
        return _build_additional_title_str(title)
    
    if isinstance(title, list):
        return _build_additional_title_list(title)
    
    return _build_additional_title_object(title)


def _build_additional_title_str(title):
    """
    Builds additional title represented by `before` and `after` parameters of ``write_exception_sync``.
    
    > This function is a type specific version called by ``_build_additional_title``.
    
    Parameters
    ----------
    title : `str`
        The title to put before or after an exception traceback.
    
    Returns
    -------
    built_title : `None`, `str`
    """
    if not title:
        return None
    
    if title.endswith('\n'):
        return title
        
    return title + '\n'


def _build_additional_title_list(title_parts):
    """
    Builds additional title represented by `before` and `after` parameters of ``write_exception_sync``.
    
    > This function is a type specific version called by ``_build_additional_title``.
    
    Parameters
    ----------
    title_parts : `list` of `str`
        The title to put before or after an exception traceback.
    
    Returns
    -------
    built_title : `None`, `str`
    """
    title_elements = []
    
    for element in title_parts:
        if not isinstance(element, str):
            element = repr(element)
        
        if element:
            title_elements.append(element)
    
    if not title_elements:
        return None
    
    if not title_elements[-1].endswith('\n'):
        title_elements.append('\n')
    
    return ''.join(title_elements)


def _build_additional_title_object(title_object):
    """
    Builds additional title represented by `before` and `after` parameters of ``write_exception_sync``.
    
    > This function is a type specific version called by ``_build_additional_title``.
    
    Parameters
    ----------
    title_object : `object`
        The title to put before or after an exception traceback.
    
    Returns
    -------
    built_title : `None`, `str`
    """
    return _build_additional_title_str(repr(title_object))


def write_exception_sync(exception, before=None, after=None, file=None, *, filter=None, highlighter=None):
    """
    Writes the given exception's traceback.
    
    Parameters
    ----------
    exception : ``BaseException``
        The exception to render.
    
    before : `None`, `str`, `list` of `str` = `None`, Optional
        Any content, what should go before the exception's traceback.
        
        If given as `str`, or if `list`, then the last element of it should end with linebreak.
    
    after : `None`, `str`, `list` of `str` = `None`, Optional
        Any content, what should go after the exception's traceback.
        
        If given as `str`, or if `list`, then the last element of it should end with linebreak.
    
    file : `None`, `I/O stream` = `None`, Optional
        The file to print the stack to. Defaults to `sys.stderr`.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    """
    
    before = _build_additional_title(before)
    after = _build_additional_title(after)
    
    if (file is None) and (highlighter is None):
        highlighter = get_default_trace_writer_highlighter()
    
    extend = []
    
    if (before is not None):
        _add_typed_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_ADDITIONAL_BEFORE,
            before,
            extend,
            highlighter,
        )
    
    render_exception_into(exception, extend, filter=filter, highlighter=highlighter)
    
    if (after is not None):
        _add_typed_part_into(
            HIGHLIGHT_TOKEN_TYPES.TOKEN_TYPE_TRACE_TITLE_ADDITIONAL_AFTER,
            after,
            extend,
            highlighter,
        )
    
    if file is None:
        # ignore exception cases
        file = sys.stderr
        
        # On shutdown `sys.stderr` can be set to `None`
        if (file is None):
            return
    
    file.write(''.join(extend))
    file.flush()


@export
def write_exception_async(exception, before=None, after=None, file=None, *, filter=None, highlighter=None, loop=None):
    """
    Writes the given exception's traceback asynchronously.
    
    Parameters
    ----------
    exception : ``BaseException``
        The exception to render.
    
    before : `None`, `str`, `list` of `str` = `None`, Optional
        Any content, what should go before the exception's traceback.
        
        If given as `str`, or if `list`, then the last element of it should end with linebreak.
    
    after : `None`, `str`, `list` of `str` = `None`, Optional
        Any content, what should go after the exception's traceback.
        
        If given as `str`, or if `list`, then the last element of it should end with linebreak.

    file : `None`, `I/O stream` = `None`, Optional
        The file to print the stack to. Defaults to `sys.stderr`.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    
    loop : `None`, ``EventThread`` = `None`, Optional (Keyword only)
        The event loop to use.
    
    Returns
    -------
    future : ``Future``
        Returns a future, what can be awaited to wait for the rendering to be done.
    """
    if loop is None:
        loop = get_event_loop()
    
    future = loop.run_in_executor(
        alchemy_incendiary(
            write_exception_sync,
            (exception, before, after, file),
            {'filter': filter, 'highlighter': highlighter},
        )
    )
    assert future.__silence__() is None
    return future


@export
def write_exception_maybe_async(exception, before=None, after=None, file=None, *, filter=None, highlighter=None):
    """
    Writes the given exception's traceback on a non-async thread. If called from an async thread will use an executor.
    If called from a sync thread, will block instead.
    
    Parameters
    ----------
    exception : ``BaseException``
        The exception to render.
    
    before : `None`, `str`, `list` of `str` = `None`, Optional
        Any content, what should go before the exception's traceback.
        
        If given as `str`, or if `list`, then the last element of it should end with linebreak.
    
    after : `None`, `str`, `list` of `str` = `None`, Optional
        Any content, what should go after the exception's traceback.
        
        If given as `str`, or if `list`, then the last element of it should end with linebreak.

    file : `None`, `I/O stream` = `None`, Optional
        The file to print the stack to. Defaults to `sys.stderr`.
    
    filter : `None`, `callable` = `None`, Optional (Keyword only)
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
        Formatter storing highlighting details.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        write_exception_async(exception, before, after, file, filter=filter, highlighter=highlighter, loop=local_thread)
    else:
        write_exception_sync(exception, before, after, file, filter=filter, highlighter=highlighter)


_DEFAULT_TRACE_WRITER_HIGHLIGHTER = None


def get_default_trace_writer_highlighter():
    """
    Returns the default highlighter for trace writer functions.
    
    Returns
    -------
    highlighter : `None`, ``HighlightFormatterContext``
    """
    return _DEFAULT_TRACE_WRITER_HIGHLIGHTER


@call
def set_default_trace_writer_highlighter():
    """
    Re-sets the default highlighter for trace writer functions.
    """
    global _DEFAULT_TRACE_WRITER_HIGHLIGHTER
    
    if PLATFORM != 'linux':
        highlighter = None
    else:
        try:
            get_terminal_size()
        except OSError:
            # If the os not supports this operation, we wont highlight
            highlighter = None
        else:
            highlighter = DEFAULT_ANSI_HIGHLIGHTER
    
    _DEFAULT_TRACE_WRITER_HIGHLIGHTER = highlighter


def set_trace_writer_highlighter(highlighter):
    """
    Sets highlighter for trace writer functions.
    
    Parameters
    ----------
    highlighter : `None`, ``HighlightFormatterContext``
        The highlighter to set.
    
    Raises
    ------
    TypeError
        - If `highlighter`'s type is incorrect.
    """
    global _DEFAULT_TRACE_WRITER_HIGHLIGHTER
    
    if (highlighter is not None) and (not isinstance(highlighter, HighlightFormatterContext)):
        raise TypeError(
            f'`highlighter` can be `None`, `{HighlightFormatterContext.__name__}, got '
            f'{highlighter.__class__.__name__}; {highlighter!r}'
        )
    
    _DEFAULT_TRACE_WRITER_HIGHLIGHTER = highlighter


EXCEPTION_MESSAGE_TITLE_STANDALONE = 'Ignoring occurred exception.\n'
EXCEPTION_MESSAGE_TITLE_BASE = 'Ignoring occurred exception at: '


class ExceptionWriterContextManager:
    """
    Exception catcher which can be used to catch and write occurred exceptions captured by it!
    
    Examples
    --------
    ```py
    from scarletio import catching
    
    def i_raise():
        raise ValueError
    
    with catching(ValueError, location="right here"):
        i_raise()
    ```
    
    Attributes
    ----------
    exclude : `None`, ``BaseException``, `tuple` of `BaseException`
        The exception types to not catch.
        
        > By default ignores `GeneratorExit` if inside of a task.
        
    filter : `None`, `callable`
        Additional filter to check whether a frame should be shown.
        
        Called with 4 parameters:
        
        +---------------+-----------+---------------------------------------------------------------+
        | Name          | Type      | Description                                                   |
        +===============+===========+===============================================================+
        | file_name     | `str`     | The frame's file's name.                                      |
        +---------------+-----------+---------------------------------------------------------------+
        | name          | `str`     | The name of the function.                                     |
        +---------------+-----------+---------------------------------------------------------------+
        | line_number   | `int`     | The line's index of the file where the exception occurred.    |
        +---------------+-----------+---------------------------------------------------------------+
        | line          | `str`     | The line of the file.                                         |
        +---------------+-----------+---------------------------------------------------------------+
    
    highlighter : `None`, ``HighlightFormatterContext``
        Formatter storing highlighting details.
    
    include : `None`, ``BaseException``, `tuple` of `BaseException`
        The exception types to catch.
        
        > By default catches everything.
    
    location : `None`, `str`
        The place where the exception might be occurring.
    
    """
    __slots__ = ('exclude', 'filter', 'highlighter', 'include', 'location')
    
    def __new__(cls, include=None, exclude=None, *, location=None, filter=None, highlighter=None):
        """
        Creates a new exception catcher.
        
        Parameters
        ----------
        include : `None`, ``BaseException``, `tuple` of `BaseException` = `None`, Optional
            The exception types to catch.
            
            > By default allows all exceptions.
        
        exclude : `None`, ``BaseException``, `tuple` of `BaseException` = `None`, Optional 
            The exception types to not catch.
            
            > By default excludes only `GeneratorExit` if inside of a task.
        
        location : `None`, `str` = `None`, Optional (Keyword only)
            The place where the exception might be occurring.
            
        filter : `None`, `callable` = `None`, Optional (Keyword only)
            Additional filter to check whether a frame should be shown.
            
            Called with 4 parameters:
            
            +---------------+-----------+---------------------------------------------------------------+
            | Name          | Type      | Description                                                   |
            +===============+===========+===============================================================+
            | file_name     | `str`     | The frame's file's name.                                      |
            +---------------+-----------+---------------------------------------------------------------+
            | name          | `str`     | The name of the function.                                     |
            +---------------+-----------+---------------------------------------------------------------+
            | line_number   | `int`     | The line's index of the file where the exception occurred.    |
            +---------------+-----------+---------------------------------------------------------------+
            | line          | `str`     | The line of the file.                                         |
            +---------------+-----------+---------------------------------------------------------------+
        
        highlighter : `None`, ``HighlightFormatterContext`` = `None`, Optional (Keyword only)
            Formatter storing highlighting details.
        """
        self = object.__new__(cls)
        self.exclude = exclude
        self.filter = filter
        self.highlighter = highlighter
        self.include = include
        self.location = location
        return self
    
    
    def __repr__(self):
        """Returns the exception catcher's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # location
        location = self.location
        if (location is not None):
            repr_parts.append(' location=')
            repr_parts.append(repr(location))
            
            field_added = True
        else:
            field_added = False
        
        # include
        include = self.include
        if (include is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' include=')
            repr_parts.append(repr(include))
        
        # exclude
        exclude = self.exclude
        if (exclude is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' exclude=')
            repr_parts.append(repr(exclude))
        
        # filter
        filter = self.filter
        if (filter is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' filter=')
            repr_parts.append(repr(filter))
        
        # highlighter
        highlighter = self.highlighter
        if (highlighter is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' highlighter=')
            repr_parts.append(repr(highlighter))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __enter__(self):
        """
        Enters the catching block as an synchronous context manager.
        
        Returns
        -------
        self : ``catch`` 
        """
        return self
    
    
    async def __aenter__(self):
        """
        Enters the catching block as an asynchronous context manager.
        
        This method is a coroutine.
        
        Returns
        -------
        self : ``catch`` 
        """
        return self
    
    
    def __exit__(self, exception_type, exception, exception_traceback):
        """
        Exits the context manager.
        
        Parameters
        ----------
        exception_type : `None`, `type<BaseException>`
            The occurred exception's type if any.
        
        exception : `None`, `BaseException`
            The occurred exception if any.
        
        exception_traceback : `None`, `TracebackType`
            the exception's traceback if any.
        
        Returns
        -------
        captured : `bool`
            Whether the exception was captured.
        """
        if not self._should_capture(exception, False):
            return False
        
        write_exception_maybe_async(
            exception,
            before = self._get_location_message(),
            filter = self.filter,
            highlighter = self.highlighter,
        )
        return True
    
    
    async def __aexit__(self, exception_type, exception, exception_traceback):
        """
        Exits the context manager.
        
        Parameters
        ----------
        exception_type : `None`, `type<BaseException>`
            The occurred exception's type if any.
        
        exception : `None`, `BaseException`
            The occurred exception if any.
        
        exception_traceback : `None`, `TracebackType`
            the exception's traceback if any.
        
        Returns
        -------
        captured : `bool`
            Whether the exception was captured.
        """
        if not self._should_capture(exception, True):
            return False
        
        await write_exception_async(
            exception,
            before = self._get_location_message(),
            filter = self.filter,
            highlighter = self.highlighter,
        )
        return True
    
    
    def _should_capture(self, exception, asynchronous):
        """
        Returns whether the given exception should be captured.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            The exception to check.
        asynchronous : `bool`
            Whether we are at an exit of an asynchronous context manager.
        
        Returns
        -------
        should_capture : `bool`
        """
        # If we have no exception we should not capture.
        if exception is None:
            return False
        
        if not asynchronous:
            thread = current_thread()
            if isinstance(thread, EventThread) and (thread.current_task is not None):
                asynchronous = True
            
            else:
                asynchronous = False
        
        if asynchronous:
            # If we are inside of a task, we ignore `GeneratorExit`
            if isinstance(exception, GeneratorExit):
                return False
        
        # If we exclude the exception we should not capture
        exclude = self.exclude
        if (exclude is not None) and isinstance(exception, exclude):
            return False
        
        # If we not include the exception we should not capture.
        include = self.include
        if (include is not None) and (not isinstance(exception, include)):
            return False
        
        return True
    
    
    def _get_location_message(self):
        """
        Creates the location message of the context.
        
        Returns
        -------
        location_message : `str`
        """
        for location_message in self._iter_location_messages():
            if (location_message is not None):
                return location_message
        
        return EXCEPTION_MESSAGE_TITLE_STANDALONE
    
    
    def _iter_location_messages(self):
        """
        Iterates over the location messages.
        
        This method is an iterable generator.
        
        Yields
        ------
        location_message : `None`, `str`
        """
        yield self._create_location_message_from_location()
        yield self._create_location_message_from_task_name()
        yield self._create_location_message_from_tracing()
    
    
    def _create_location_message_from_location(self):
        """
        Creates location message using ``.location``
        
        Returns
        -------
        location_message : `None`, `str`
        """
        location = self.location
        if (location is not None):
            return f'{EXCEPTION_MESSAGE_TITLE_BASE}{location}\n'
    
    
    def _create_location_message_from_task_name(self):
        """
        Creates location message from the current task's qualname.
        
        Returns
        -------
        location_message : `None`, `str`
        """
        thread = current_thread()
        if isinstance(thread, EventThread):
            current_task = thread.current_task
            if (current_task is not None):
                return f'{EXCEPTION_MESSAGE_TITLE_BASE}{current_task.qualname}\n'
    
    
    def _create_location_message_from_tracing(self):
        """
        Creates location message by tracing back to the first non-local frame.
        
        Returns
        -------
        location_message : `None`, `str`
        """
        frame = get_frame()
        while True:
            if frame is None:
                return
            
            if frame.f_code.co_filename != __file__:
                break
            
            frame = frame.f_back
            continue
        
        return EXCEPTION_MESSAGE_TITLE_BASE + frame.f_code.co_name


# Blackmagic

def __system_exception_hook__(exception_type, exception, exception_traceback):
    """
    Displays the given exception. Used to replace the system default.

    Parameters
    ----------
    exception_type : `None`, `type<BaseException>`
        The occurred exception's type if any.
    
    exception : `None`, `BaseException`
        The occurred exception if any.
    
    exception_traceback : `None`, `TracebackType`
        the exception's traceback if any.
    """
    if (exception is not None):
        write_exception_sync(exception)


def __threading_exception_hook__(threading_exception_hook_parameters):
    """
    Displays the given exception. Used to replace threading default.
    
    Parameters
    ----------
    threading_exception_hook_parameters : `tuple`
        Threading exception hook parameters, which are:
        - exception_type
        - exception
        - exception_traceback
        - thread
    """
    exception_type, exception, exception_traceback, thread = threading_exception_hook_parameters
    if (exception is not None):
        if thread is None:
            thread_name = str(get_ident())
        else:
            thread_name = thread.name
        
        write_exception_sync(exception, before=f'Exception in thread: {thread_name}\n')


@call
def set_exception_hooks():
    """
    Replaces python's exception hooks.
    """
    attribute_name = f'{"except"}{"hook"}' # We do this, so we do not get typo error
    
    setattr(sys, attribute_name, __system_exception_hook__)
    setattr(threading, attribute_name, __threading_exception_hook__)



__all__ = ('WebSocketCommonProtocol',)

import hashlib
from base64 import b64decode, b64encode
from binascii import Error as BinasciiError
from email.utils import formatdate

from ..core import AsyncQueue, CancelledError, Future, Lock, Task, write_exception_async
from ..utils import IgnoreCaseMultiValueDictionary, is_coroutine
from ..web_common import AbortHandshake, InvalidHandshake, InvalidOrigin, InvalidUpgrade, PayloadError
from ..web_common.header_building_and_parsing import (
    build_extensions, parse_connections, parse_extensions, parse_subprotocols, parse_upgrades
)
from ..web_common.headers import (
    CONNECTION, CONTENT_LENGTH, CONTENT_TYPE, DATE, ORIGIN, SEC_WEBSOCKET_ACCEPT, SEC_WEBSOCKET_EXTENSIONS,
    SEC_WEBSOCKET_KEY, SEC_WEBSOCKET_PROTOCOL, SEC_WEBSOCKET_VERSION, SERVER, UPGRADE
)

from .websocket_common_protocol import (
    BAD_REQUEST, FORBIDDEN, INTERNAL_SERVER_ERROR, SERVICE_UNAVAILABLE, SWITCHING_PROTOCOLS, UPGRADE_REQUIRED,
    WEBSOCKET_KEY, WebSocketCommonProtocol
)


class WebSocketServerProtocol(WebSocketCommonProtocol):
    """
    Asynchronous server side websocket protocol implementation.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_waiter : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `Any`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    _drain_waiter : `None`, ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    _drain_lock : ``Lock``
        Asynchronous lock to ensure, that only `1` frame is written in `1` time.
    close_code : `int`
        The websocket's close code if applicable. Defaults to `0`.
    close_connection_task : `None`, ``Task`` of ``.close_connection``
        A task, what is present meanwhile the websocket is closing to avoid race condition.
    close_timeout : `float`
        The maximal duration in seconds what is waited for response after close frame is sent. Defaults to `10.0`.
    close_reason : `None`, `str`
        The reason, why the websocket was closed. Set only after the websocket is closed. Close reason might not be
        received tho.
    connection_lost_waiter : ``Future``
        A future, what's result is set as `None`, when the connection is closed. Used to wait for close frames.
        
        ``shield`` it if using from outside.
    extensions : `None` or (`list` of `Any`)
        Web socket extensions. Defaults to `None`, if there is not any.
    host : `str`
        The respective server's address to connect to.
    max_queue : `None`, `int`
        Max queue size of ``.messages``. If a new payload is added to a full queue, the oldest element of it is removed.
         Defaults to `None`.
    max_size : `int`
        Max payload size to receive. If a payload exceeds it, ``PayloadError`` is raised. Defaults to `67108864` bytes.
    messages : ``AsyncQueue``
        An asynchronous queue of the received messages.
    is_ssl : `bool`
        Whether the connection is secure. Defaults to `False`.
    pings : `OrderedDict` of (`bytes`, ``Future``) items
        An ordered dictionary of ping payloads and of their waiter futures.
    port : `int`
        The respective server's port to connect to.
    state : `str`
        The websocket's state.
        
        Can be set as one of the following values:
        
        +-------------------------------+-------------------+
        | Respective name               | Value             |
        +===============================+===================+
        | WEBSOCKET_STATE_CONNECTING    | `1`               |
        +-------------------------------+-------------------+
        | WEBSOCKET_STATE_OPEN          | `2`               |
        +-------------------------------+-------------------+
        | WEBSOCKET_STATE_CLOSING       | `3`               |
        +-------------------------------+-------------------+
        | WEBSOCKET_STATE_CLOSED        | `4`               |
        +-------------------------------+-------------------+
        
        Note, that state == compared by memory address and not by value.
    subprotocol : `None`, `str`
        Chosen subprotocol at handshake. Defaults to `None` and might be set as `str`. Chosen from the available
        subprotocols by their priority order.
    transfer_data_exception : `None`, `BaseException``
        Exception catched meanwhile processing received data.
    transfer_data_task : `None`, ``Task`` of ``.transfer_data``
        Data receiving task.
    available_extensions : `None` or (`list` of `Any`)
        Available websocket extensions. Defaults to `None`.
        
        Each websocket extension should have the following `4` attributes / methods:
        - `name`: `str`. The extension's name.
        - `request_params` : `list` of `tuple` (`str`, `str`). Additional header parameters of the extension.
        - `decode` : `callable`. Decoder method, what processes a received websocket frame. Should accept `2`
            parameters: The respective websocket ``WebSocketFrame``, and the ˙max_size` as `int`, what describes the
            maximal size of a received frame. If it is passed, ``PayloadError`` is raised.
        - `encode` : `callable`. Encoder method, what processes the websocket frames to send. Should accept `1`
            parameter, the respective websocket ``WebSocketFrame``.
    available_subprotocols : `None` or (`list` of `str`)
        A list of supported subprotocols in order of decreasing preference.
    extra_response_headers : ``IgnoreCaseMultiValueDictionary``, `dict-like` with (`str`, `str`) items
        Extra response headers.
    handler : `async-callable`
        An asynchronous callable, what will handle a websocket connection.
        
        Should be given as an `async-callable` accepting `1` parameter the respective asynchronous server side
        websocket protocol implementations.
    handler_task : `None`, ``Task`` of ``.lifetime_handler``
        Handles the connected websocket meanwhile it is alive.
    origin : `None`, `str`
        Value of the Origin header.
    request_processor : `None`, `callable`
        An optionally asynchronous callable, what processes the initial requests from the potential clients.
        
        Should accept the following parameters:
        - `path` : `str`. The requested path.
        - `request_headers` : ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`). The request's headers.
        
        The `request_processor` on accepted request should return `None`, otherwise a `tuple` of
        ``AbortHandshake`` parameters.
    server : ``WSServer``
        The owner websocket server instance.
    subprotocol_selector `None`, `callable`
        User hook to select subprotocols. Should accept the following parameters:
        - `parsed_header_subprotocols` : `list` of `str`. The subprotocols supported by the client.
        - `available_subprotocols` : `list` of `str`. The subprotocols supported by the server.
    request : `None`, ``RawRequestMessage``
        The received http request if applicable.
    response_headers : `None`, ``IgnoreCaseMultiValueDictionary`` of (`str`, `str`) items
        The server websocket's response's headers if applicable.
    """
    is_client = False
    
    __slots__ = (
        'available_extensions', 'available_subprotocols', 'extra_response_headers', 'handler', 'handler_task',
        'origin', 'origin', 'request_processor', 'server', 'subprotocol_selector', 'request', 'response_headers'
    )
    
    def __new__(cls, server):
        """
        Creates a new ``WebSocketServerProtocol`` with the given parameters.
        
        This method is usually wrapped into a partial function.
        
        Parameters
        ----------
        server : ``WSServer``
            The parent websocket server.
        """
        handler, host, port, is_ssl, origin, available_extensions, available_subprotocols , extra_response_headers, \
        request_processor, subprotocol_selector, websocket_kwargs = server.protocol_parameters
        
        self = WebSocketCommonProtocol.__new__(cls, server.loop, host, port, is_ssl=is_ssl, **websocket_kwargs)
        self.handler = handler
        self.server = server
        self.origin = origin
        self.available_extensions = available_extensions
        self.available_subprotocols = available_subprotocols
        self.extra_response_headers = extra_response_headers
        self.request_processor = request_processor
        self.subprotocol_selector = subprotocol_selector
        self.handler_task = None
        
        self.request = None
        self.response_headers = None
        self.origin = None
        
        return self
    
    def connection_made(self, transport):
        """
        Called when a connection is made.
        
        Parameters
        ----------
        transport : `Any`
            Asynchronous transport implementation, what calls the protocol's ``.data_received`` when data is
            received.
        """
        WebSocketCommonProtocol.connection_made(self, transport)
        self.server.register(self)
        self.handler_task = Task(self.lifetime_handler(), self._loop)
    
    
    async def lifetime_handler(self):
        """
        The asynchronous websocket protocol's main "lifetime" task.
        
        This method is a coroutine.
        """
        try:
            # handshake returns True if it succeeded
            if not (await self.handshake()):
                return
            
            try:
                await self.handler(self)
            except (GeneratorExit, CancelledError):
                raise
            
            except BaseException as err:
                write_exception_async(
                    err,
                    [
                        'Unhandled exception occurred at',
                        self.__class__.__name__,
                        '.lifetime_handler meanwhile running: ',
                        repr(self.handler),
                        '\n',
                    ],
                    loop = self._loop,
                )
                return
            
            await self.close()
        except:
            # We will let Task.__del__ to render the exception...
            
            transport = self._transport
            if transport is None:
                raise
                
            transport.close()
            transport.abort()
            raise
        
        finally:
            self.handler_task = None
            self.server.unregister(self)
    
    
    async def handshake(self):
        """
        Handles a received websocket connect request.
        
        This method is a coroutine.
        
        Returns
        -------
        handshake_succeeded : `bool`
            If the websocket handshake succeeded and starting, it's handler can begin, returns `True`.
        """
        try:
            self.request = request = await self.set_payload_reader(self._read_http_request())
            
            request_headers = request.headers
            if self.server.is_serving():
                path = request.path
                
                request_processor = self.request_processor
                if request_processor is None:
                    early_response = None
                else:
                    early_response = request_processor(path, request_headers)
                    if is_coroutine(early_response):
                        early_response = await early_response
                
                if (early_response is not None):
                    raise AbortHandshake(
                        *early_response,
                        request = request,
                    )
                
            else:
                raise AbortHandshake(
                    SERVICE_UNAVAILABLE,
                    'Server is shutting down.',
                    request = request,
                )
            
            connections = []
            connection_headers = request_headers.get_all(CONNECTION)
            if (connection_headers is not None):
                for connection_header in connection_headers:
                    connections.extend(parse_connections(connection_header))
        
            if not any(value.lower() == 'upgrade' for value in connections):
                raise InvalidUpgrade(
                    f'Invalid connection, no upgrade found, got {connections!r}.'
                )
            
            upgrade = []
            upgrade_headers = request_headers.get_all(UPGRADE)
            if (upgrade_headers is not None):
                for upgrade_header in upgrade_headers:
                    upgrade.extend(parse_upgrades(upgrade_header))
            
            if len(upgrade) != 1 and upgrade[0].lower() != 'websocket': # ignore case
                raise InvalidUpgrade(
                    f'Expected \'WebSocket\' for \'Upgrade\', but got {upgrade!r}.'
                )
            
            received_keys = request_headers.get_all(SEC_WEBSOCKET_KEY)
            if received_keys is None:
                raise InvalidHandshake(
                    f'Missing {SEC_WEBSOCKET_KEY!r} from headers',
                    request = request,
                )
            
            if len(received_keys) > 1:
                raise InvalidHandshake(
                    f'Multiple {SEC_WEBSOCKET_KEY!r} values at headers',
                    request = request,
                )
            
            key = received_keys[0]
        
            try:
                raw_key = b64decode(key.encode(), validate=True)
            except BinasciiError:
                raise InvalidHandshake(
                    f'Invalid {SEC_WEBSOCKET_KEY!r}: {key!r}.',
                    request = request)
            
            if len(raw_key) != 16:
                raise InvalidHandshake(
                    f'Invalid {SEC_WEBSOCKET_KEY!r}, should be length 16; {key!r}.',
                    request = request,
                )
            
            sw_version = request_headers.get_all(SEC_WEBSOCKET_VERSION)
            if sw_version is None:
                raise InvalidHandshake(
                    f'Missing {SEC_WEBSOCKET_VERSION!r} values at headers.',
                    request = request,
                )
            
            if len(sw_version) > 1:
                raise InvalidHandshake(
                    f'Multiple {SEC_WEBSOCKET_VERSION!r} values at headers.',
                    request = request,
                )
            
            sw_version = sw_version[0]
            if sw_version != '13':
                raise InvalidHandshake(
                    f'Invalid {SEC_WEBSOCKET_VERSION!r}: {sw_version!r}.',
                    request = request,
                )
            
            while True:
                origin = self.origin
                if (origin is None):
                    origin = None
                    break
                
                origin_headers = request_headers.get_all(ORIGIN)
                
                if (origin_headers is None):
                    raise InvalidOrigin('No origin at header.')
                
                if len(origin_headers) > 1:
                    raise InvalidOrigin('More than 1 origin at header.')
                
                origin = origin_headers[0]
                
                if origin in origin:
                    break
                    
                raise InvalidOrigin(origin)
            
            self.origin = origin
            
            while True:
                accepted_extensions = []
                
                available_extensions = self.available_extensions
                if (available_extensions is None):
                    extension_header = None
                    break
                
                extension_headers_ = request_headers.get_all(SEC_WEBSOCKET_EXTENSIONS)
                if (extension_headers_ is None):
                    extension_header = None
                    break
                
                extension_headers = []
                parsed_extension_values = []
                for extension_header_ in extension_headers_:
                    parsed_extension_values.extend(parse_extensions(extension_header_))
                
                for name, params in parsed_extension_values:
                    for extension in available_extensions:
                        # do names and params match?
                        if extension.name == name and extension.are_valid_params(params, accepted_extensions):
                            accepted_extensions.append(extension)
                            extension_headers.append((name, params))
                            break
                    else:
                        # no matching extension
                        raise InvalidHandshake(
                            f'Unsupported extension: name={name!r}, params={params!r}.',
                            request = request,
                        )
                    
                    # If we didn't break from the loop, no extension in our list matched what the client sent. The
                    # extension is declined.
                
                # Serialize extension header.
                if extension_headers:
                    extension_header = build_extensions(extension_headers)
                    break
                
                extension_header = None
                break
            
            self.extensions = accepted_extensions
            
            
            while True:
                available_subprotocols = self.available_subprotocols
                if (available_subprotocols is None):
                    selected_subprotocol = None
                    break
                    
                protocol_headers = request_headers.get_all(SEC_WEBSOCKET_PROTOCOL)
                if (protocol_headers is None):
                    selected_subprotocol = None
                    break
                
                parsed_header_subprotocols = []
                for protocol_header in protocol_headers:
                    parsed_header_subprotocols.extend(parse_subprotocols(protocol_header))
                
                subprotocol_selector = self.subprotocol_selector
                if (subprotocol_selector is not None):
                    selected_subprotocol = subprotocol_selector(parsed_header_subprotocols, available_subprotocols)
                    break
                    
                subprotocols = set(parsed_header_subprotocols)
                subprotocols.intersection_update(available_subprotocols)
                
                if not subprotocols:
                    selected_subprotocol = None
                    break
                
                lowest_priority = len(parsed_header_subprotocols) + len(available_subprotocols)
                selected_subprotocol = None
                
                for subprotocol in subprotocols:
                    priority = parsed_header_subprotocols.index(subprotocol) + available_subprotocols.index(subprotocol)
                    if priority < lowest_priority:
                        lowest_priority = priority
                        selected_subprotocol = subprotocol
                
                break
            
            self.subprotocol = selected_subprotocol
            
            response_headers = IgnoreCaseMultiValueDictionary()
    
            response_headers[UPGRADE] = 'websocket'
            response_headers[CONNECTION] = 'Upgrade'
            response_headers[SEC_WEBSOCKET_ACCEPT] = \
                b64encode(hashlib.sha1((key + WEBSOCKET_KEY).encode()).digest()).decode()
            
            if (extension_header is not None):
                response_headers[SEC_WEBSOCKET_EXTENSIONS] = extension_header
            
            if (selected_subprotocol is not None):
                response_headers[SEC_WEBSOCKET_PROTOCOL] = selected_subprotocol
            
            extra_response_headers = self.extra_response_headers
            if (extra_response_headers is not None):
                for key, value in extra_response_headers.items():
                    response_headers[key] = value
            
            response_headers.setdefault(DATE, formatdate(usegmt=True))
            response_headers.setdefault(SERVER, '')
            
            self.response_headers = response_headers
            self.write_http_response(SWITCHING_PROTOCOLS, response_headers)
            
            self.connection_open()
        
        except (GeneratorExit, CancelledError):
            raise
        
        except ConnectionError as err:
            write_exception_async(
                err,
                [
                    'Unhandled exception occurred at ',
                    self.__class__.__name__,
                    '.handshake, when handshaking:\n'
                ],
                loop = self._loop,
            )
            return False
        
        except BaseException as err:
            if isinstance(err, AbortHandshake):
                status = err.code
                headers = err.headers
                if headers is None:
                    headers = IgnoreCaseMultiValueDictionary()
                body = err.message
                if not body.endswith('\n'):
                    body = body + b'\n'
            
            elif isinstance(err, InvalidOrigin):
                status = FORBIDDEN
                headers = IgnoreCaseMultiValueDictionary()
                body = f'Failed to open a WebSocket connection: {err}.\n'.encode()
            
            elif isinstance(err, InvalidUpgrade):
                status = UPGRADE_REQUIRED
                headers = IgnoreCaseMultiValueDictionary()
                headers[UPGRADE] = 'websocket'
                body = (
                    f'Failed to open a WebSocket connection: {err}.\n\n'
                    f'You cannot access a WebSocket server directly with a browser. You need a WebSocket client.\n'
                ).encode()
            
            elif isinstance(err, InvalidHandshake):
                status = BAD_REQUEST
                headers = IgnoreCaseMultiValueDictionary()
                body = f'Failed to open a WebSocket connection: {err.message}.\n'.encode()
            
            elif isinstance(err, PayloadError):
                status = BAD_REQUEST
                headers = IgnoreCaseMultiValueDictionary()
                body = f'Invalid request body: {err}.\n'.encode()
            
            else:
                status = INTERNAL_SERVER_ERROR
                headers = IgnoreCaseMultiValueDictionary()
                body = b'Failed to open a WebSocket connection.\n'
            
            headers.setdefault(DATE, formatdate(usegmt=True))
            headers.setdefault(SERVER, '')
            headers.setdefault(CONTENT_LENGTH, repr(len(body)))
            headers.setdefault(CONTENT_TYPE, 'text/plain')
            headers.setdefault(CONNECTION, 'close')
            
            try:
                self.write_http_response(status, headers, body=body)
                self.fail_connection()
                await self.wait_for_connection_lost()
            except (GeneratorExit, CancelledError):
                raise
            
            except BaseException as err2:
                write_exception_async(
                    err2,
                    [
                        'Unhandled exception occurred at ',
                        self.__class__.__name__,
                        '.handshake, when handling an other exception;',
                        repr(err),
                        ':\n',
                    ],
                    loop = self._loop,
                )
            return False
        
        return True



__all__ = ('FutureAsyncWrapper', )

import reprlib
from types import MethodType

from ...utils import copy_docs, export, ignore_frame, set_docs
from ...utils.trace import format_callback

from ..exceptions import InvalidStateError

from .future import (
    FUTURE_STATE_CANCELLED, FUTURE_STATE_FINISHED, FUTURE_STATE_PENDING, FUTURE_STATE_RETRIEVED, Future,
    get_exception_short_representation, get_future_state_name
)


ignore_frame(__spec__.origin, 'get_result', 'raise exception',)
ignore_frame(__spec__.origin, '__iter__', 'yield self',)

@export
class FutureAsyncWrapper(Future):
    """
    Async wrapper for ``Future``-s enabling them to be awaited from an another event loop.
    
    Attributes
    ----------
    _blocking : `bool`
        Whether the future is already being awaited, so it blocks the respective coroutine.
    _callbacks : `list` of `callable`
        The callbacks of the future, which are queued up on the respective event loop to be called, when the future is
        finished. These callback should accept `1` parameter, the future itself.
        
        Note, if the future is already done, then the newly added callbacks are queued up instantly on the respective
        event loop to be called.
    
    _exception : `None`, `BaseException`
        The exception set to the future as it's result. Defaults to `None`.
    _future : `None`, ``Future``
        The waited future. If the future's state == modified by the sync wrapper, then ``._future`` is set as `None`,
        to not retrieve the result again.
    _loop : ``EventThread``
        The loop to what the async future wrapper is bound to.
    _result : `None`, `Any`
        The result of the future. Defaults to `None`.
    _state : `str`
        The state of the future.
        
        Can be set as one of the following:
        
        +---------------------------+-----------+
        | Respective name           | Value     |
        +===========================+===========+
        | FUTURE_STATE_PENDING      | `0`       |
        +---------------------------+-----------+
        | FUTURE_STATE_CANCELLED    | `1`       |
        +---------------------------+-----------+
        | FUTURE_STATE_FINISHED     | `2`       |
        +---------------------------+-----------+
        | FUTURE_STATE_RETRIEVED    | `3`       |
        +---------------------------+-----------+
        
        Note, that states are checked by memory address and not by equality. Also ``FUTURE_STATE_RETRIEVED`` is used only if
        `__debug__` is set as `True`.
    """
    __slots__ = ('_blocking', '_callbacks', '_exception', '_future', '_loop', '_result', '_state',)

    def __new__(cls, future, loop):
        """
        Creates a new ``FutureAsyncWrapper`` object bound to the given `loop` and `future`.
        
        If the given `future` is an ``FutureAsyncWrapper``, then will wrap it's future instead.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop from where the created wrapper futures can be awaited.
        future : ``Future``, ``FutureAsyncWrapper``
            The future to wrap.
        
        Returns
        -------
        self : ``Future``, ``FutureAsyncWrapper``
            If the given `future` is bound to the given loop, returns the `future` itself instead of async wrapping it.
        """
        if future._loop is loop:
            return future
        
        if isinstance(future, cls):
            future = future._future
            if future._loop is loop:
                return future
        
        self = object.__new__(cls)
        self._future = future
        self._loop = loop
        self._state = FUTURE_STATE_PENDING
        self._result = None
        self._exception = None
        
        self._callbacks = []
        self._blocking = False
        
        loop = future._loop
        loop.call_soon(future.add_done_callback, self._done_callback)
        loop.wake_up()
        
        return self
    
    
    def __call__(self, future):
        """
        By calling a ``FutureAsyncWrapper`` you can make it bound to an other future.
        
        Parameters
        ----------
        future : ``Future``
            The future on what you would want to wait from the other event loop.
        
        Returns
        -------
        self : ``FutureAsyncWrapper``
        """
        old_future = self._future
        if old_future is not None:
            loop = old_future._loop
            loop.call_soon(self._remove_callback, old_future)
            loop.wake_up()
        
        self._future = future
        self._state = FUTURE_STATE_PENDING
        self._result = None
        self._exception = None
        self._blocking = False
        
        loop = future._loop
        loop.call_soon(future.add_done_callback, self._done_callback)
        loop.wake_up()
    
    
    def __repr__(self):
        """Returns the future async wrapper's representation."""
        repr_parts = ['<', self.__class__.__name__, ' ']
        
        state = self._state
        repr_parts.append(get_future_state_name(state))
        
        if state >= FUTURE_STATE_FINISHED:
            exception = self._exception
            if exception is None:
                repr_parts.append(', result=')
                repr_parts.append(reprlib.repr(self.result))
            else:
                repr_parts.append(', exception=')
                repr_parts.append(get_exception_short_representation(exception))
        
        callbacks = self._callbacks
        limit = len(callbacks)
        if limit:
            repr_parts.append(', callbacks=[')
            index = 0
            while True:
                callback = callbacks[index]
                repr_parts.append(format_callback(callback))
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        future = self._future
        if future is not None:
            # we do not want to repr it, keep it thread safe
            repr_parts.append(', future=')
            repr_parts.append(future.__class__.__name__)
            repr_parts.append('(...)')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    if __debug__:
        @copy_docs(Future.cancel)
        def cancel(self):
            state = self._state
            if state != FUTURE_STATE_PENDING:
                if state == FUTURE_STATE_FINISHED:
                    self._state = FUTURE_STATE_RETRIEVED
                
                return 0
            
            future = self._future
            if future is None:
                self._state = FUTURE_STATE_CANCELLED
                self._loop._schedule_callbacks(self)
                return 1
            
            loop = future._loop
            loop.call_soon(future.cancel)
            loop.wake_up()
            return 1
        
    else:
        @copy_docs(Future.cancel)
        def cancel(self):
            if self._state != FUTURE_STATE_PENDING:
                return 0
            
            future = self._future
            if future is None:
                self._state = FUTURE_STATE_CANCELLED
                self._loop._schedule_callbacks(self)
                return 1
            
            loop = future._loop
            loop.call_soon(future.cancel)
            loop.wake_up()
            return 1
    
    
    if __debug__:
        def _done_callback(self, future):
            if self._future is not future:
                return
            
            state = future._state
            if state == FUTURE_STATE_FINISHED:
                future._state = FUTURE_STATE_RETRIEVED
            elif state == FUTURE_STATE_RETRIEVED:
                state = FUTURE_STATE_FINISHED
            
            loop = self._loop
            loop.call_soon(self.__class__._done_callback_re, self, state, future._result, future._exception)
            loop.wake_up()
    
    else:
        def _done_callback(self, future):
            if (self._future is not future):
                return
            
            loop = self._loop
            loop.call_soon(self.__class__._done_callback_re, self, future._state, future._result, future._exception)
            loop.wake_up()
    
    
    set_docs(_done_callback,
        """
        Callback added to the waited future to retrieve it's result.
        
        Parameters
        ----------
        future : ``Future``
            The waited future.
        
        Notes
        -----
        If a different future is given as parameter than the currently waited one, then wont do anything.
        
        If `_debug__` is set as `True`, then this callback also marks the waited future as retrieved.
        """
    )
    
    def _done_callback_re(self, state, result, exception):
        """
        Function queued up on the wrapper future's loop, by ``._done_callback``, marking this future as done.
        
        Parameters
        ----------
        state : `str`
            The future's new state.
        result : `Any`
            The future's new result.
        exception : `Any`
            The future's new exception.
        """
        self._state = state
        self._result = result
        self._exception = exception
        
        self._loop._schedule_callbacks(self)
    
    
    def _remove_callback(self, future):
        """
        Removes ``._done_callback`` callback from the given `future`'s callbacks.
        
        Parameters
        ----------
        future : ``Future``
            The future, from what's callbacks the own one will be removed.
        """
        callbacks = future._callbacks
        if callbacks:
            for index in range(len(callbacks)):
                callback = callbacks[index]
                if (type(callback) is MethodType) and (callback.__self__ is self):
                    del callbacks[index]
                    break
    
    
    def _set_future_result(self, future, result):
        """
        Sets the given `result` as the `future`'s result if applicable.
        
        This method is put on the respective event loop to be called, by ``.set_result`` or by
        ``.set_result_if_pending``, if the future's result can be set.
        
        Parameters
        ----------
        future : ``Future``
            The future to set the result to.
        result : `Any`
            The result to set as the future's.
        """
        try:
            future.set_result(result)
        except (RuntimeError, InvalidStateError): # the future does not support this operation
            pass
        
        loop = self._loop
        loop.call_soon(self.__class__._set_future_any_re, self, future)
        loop.wake_up()
    
    
    def _set_future_any_re(self, future):
        """
        If self is still waiting on the given future, marks self as retrieved and removes `._future`, so further results
        wont be retrieved.
        
        Parameters
        ----------
        future : ``Future``.
        """
        if self._future is future:
            self._state = FUTURE_STATE_RETRIEVED
            self._future = None
    
    
    @copy_docs(Future.set_result)
    def set_result(self, result):
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(self, 'set_result')
        
        future = self._future
        
        if future is None:
            self._result = result
            self._state = FUTURE_STATE_FINISHED
            self._loop._schedule_callbacks(self)
            return
        
        loop = future._loop
        loop.call_soon(self._set_future_result, future, result)
        loop.wake_up()
    
    
    @copy_docs(Future.set_result_if_pending)
    def set_result_if_pending(self, result):
        if self._state != FUTURE_STATE_PENDING:
            return 1
        
        future = self._future
        
        if future is None:
            self._result = result
            self._state = FUTURE_STATE_FINISHED
            self._loop._schedule_callbacks(self)
            return 2
        
        loop = future._loop
        loop.call_soon(self._set_future_result, future, result)
        loop.wake_up()
        return 1
    
    
    def _set_future_exception(self, future, exception):
        """
        Sets the given `exception` as the `future`'s exception if applicable.
        
        This method is put on the respective event loop to be called, by ``.set_exception`` or by
        ``.set_exception_if_pending``, if the future's result can be set.
        
        Parameters
        ----------
        future : ``Future``
            The future to set the result to.
        exception : `Any`
            The exception to set as the future's.
        """
        try:
            future.set_exception(exception)
        except (RuntimeError, InvalidStateError): #the future does not supports this operation
            pass
        
        loop = self._loop
        loop.call_soon(self._set_future_any_re, future)
        loop.wake_up()
    
    
    @copy_docs(Future.set_exception)
    def set_exception(self, exception):
        if self._state != FUTURE_STATE_PENDING:
            raise InvalidStateError(self, 'set_exception')
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        future = self._future
        if future is None:
            self._exception = exception
            self._state = FUTURE_STATE_FINISHED
            self._loop._schedule_callbacks(self)
            return
        
        loop = future._loop
        loop.call_soon(self._set_future_exception, future, exception)
        loop.wake_up()
    
    
    @copy_docs(Future.set_exception_if_pending)
    def set_exception_if_pending(self, exception):
        if self._state != FUTURE_STATE_PENDING:
            return 0
        
        if isinstance(exception, type):
            exception = exception()
        
        if isinstance(exception, StopIteration):
            raise TypeError(
                f'{exception} cannot be raised to a(n) `{self.__class__.__name__}`; {self!r}.'
            )
        
        future = self._future
        if future is None:
            self._exception = exception
            self._state = FUTURE_STATE_FINISHED
            self._loop._schedule_callbacks(self)
            return 2
        
        loop = future._loop
        loop.call_soon(self._set_future_exception, future, exception)
        loop.wake_up()
        return 1
    
    
    @copy_docs(Future.clear)
    def clear(self):
        future = self._future
        if future is not None:
            loop = future._loop
            loop.call_soon(self._remove_callback, future)
            loop.wake_up()
            self._future = None

        self._state = FUTURE_STATE_PENDING
        self._exception = None
        self._result = None
        self.cancel_handles()
        self._blocking = False



__all__ = ()

from ..core.top_level import write_exception_async


class Connection:
    """
    Reusable connection used by connectors to remember their details.
    
    Attributes
    ----------
    callbacks : `list` of `callable`
        Callable-s to run when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed. They should accept no
        parameters.
    connector : ``ConnectorBase``
        The respective connector of the connection.
    key : ``ConnectionKey``
        A key which contains information about the host.
    protocol : `None`, ``HttpReadWriteProtocol``
        The connection's actual protocol. Set as `None` if the connection is ``.close``-d, ``.release``-d or
        ``.detach``-ed.
    """
    __slots__ = ('callbacks', 'connector', 'key', 'protocol',)
    
    def __init__(self, connector, key, protocol):
        """
        Creates a new ``Connection`` with the given parameters.
        
        connector : ``ConnectorBase``
            The respective connector of the connection.
        key : `tuple` (`str`, `int`)
            The host and the port to what the connection is connected too.
        protocol : ``HttpReadWriteProtocol``
            The connection's actual protocol.
        """
        self.connector = connector
        self.key = key
        self.protocol = protocol
        self.callbacks = []
    
    def __repr__(self):
        """Returns the representation of the connection."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' to ',
            repr(self.key),
        ]
        
        if self.closed:
            repr_parts.append(' closed')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __del__(self):
        """Releases the connector if not yet released."""
        connector = self.connector
        if connector.loop.running:
            protocol = self.protocol
            if (protocol is not None):
                connector.release(self.key, protocol, should_close=True)
    
    
    @property
    def transport(self):
        """
        Returns the connection's ``.protocol``'s transport if applicable.
        
        Returns
        -------
        transport : `Any`
            Defaults to `None`.
        """
        protocol = self.protocol
        if protocol is None:
            return None
        
        return protocol.get_transport()
    
    @property
    def writer(self):
        """
        Returns the connection's ``.protocol``'s writer if applicable.
        
        Returns
        -------
        writer : `Any`
            Defaults to `None`.
        """
        protocol = self.protocol
        if protocol is None:
            return None
        
        return protocol.writer
    
    
    def add_callback(self, callback):
        """
        Adds a callback to the connection. If the connection is already closed runs it instantly.
        
        Parameters
        ----------
        callback : `callable`
            Callable to run when the connection is ``.close``-d, ``.release``-d or ``.detach``-ed. Should accept no
            parameters.
        """
        if callback is None:
            return
        
        if not self.closed:
            self.callbacks.append(callback)
            return
        
        
        try:
            callback()
        except BaseException as err:
            write_exception_async(
                err,
                [
                    'Exception occurred at ',
                    repr(self),
                    '.add_callback\nAt running ',
                    repr(callback),
                    '\n',
                ],
                loop = self.connector.loop,
            )
    
    
    def _run_callbacks(self):
        """
        Runs the added callbacks of the connection.
        """
        callbacks = self.callbacks
        while callbacks:
            callback = callbacks.pop()
            
            try:
                callback()
            except BaseException as err:
                write_exception_async(
                    err,
                    [
                        'Exception occurred at ',
                        repr(self),
                        '._run_callbacks\nAt running ',
                        repr(callback),
                        '\n',
                    ],
                    loop = self.connector.loop,
                )
    
    
    def close(self):
        """
        Closes the connection by running it's callbacks and releasing with the intent of closing.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release(self.key, protocol, should_close=True)
    
    
    def release(self):
        """
        Closes the connection by running it's callbacks and releasing it.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release(self.key, protocol, should_close=protocol.should_close())
    
    
    def detach(self):
        """
        Detaches the connection from it ``.connector``.
        
        Note, that with detaching, the connection set it's ``.protocol`` as `None`, so you should save an instance of
        it down before.
        """
        self._run_callbacks()
        
        protocol = self.protocol
        if (protocol is not None):
            self.protocol = None
            self.connector.release_acquired_protocols(self.key, protocol)
   
   
    @property
    def closed(self):
        """
        Returns whether the connection is closed.
        
        Returns
        -------
        closed : `bool`
        """
        protocol = self.protocol
        if protocol is None:
            return True
        
        if protocol.get_transport() is None:
            return True
        
        return False



__all__ = ('run', 'run_coroutine')

from threading import current_thread

from ..event_loop import EventThread
from ..traps import Task

from .event_loop import _get_event_loop_is_current_thread, _try_detect_event_loop


def run(awaitable, timeout=None):
    """
    Runs the given awaitable on an event loop. This function cannot be called from an event loop itself.
    
    Parameters
    ----------
    awaitable : `awaitable`
        The awaitable to run.
    timeout : `None`, `float` = `None`, Optional
        Timeout after the awaitable should be cancelled. Defaults to `None`.
    
    Raises
    ------
    TypeError
        If `awaitable` is not `awaitable`.
    TimeoutError
         If `awaitable` did not finish before the given `timeout` is over.
    RuntimeError
        If called from itself.
    BaseException
        Any exception raised by `awaitable`.
    """
    local_thread = current_thread()
    if isinstance(local_thread, EventThread):
        raise RuntimeError(
            f'`{local_thread.__class__.__name__}.run` should not be called from itself; thread={local_thread!r}.'
        )
    
    event_loop = _try_detect_event_loop(local_thread)
    if (event_loop is None):
        stop_event_loop_after = True
        event_loop = EventThread(daemon=True, name='scarletio.run', start_later=False)
    else:
        stop_event_loop_after = False
    
    try:
        return event_loop.run(awaitable, timeout=timeout)
    finally:
        if stop_event_loop_after:
            event_loop.stop()


def run_coroutine(coroutine, loop=None):
    """
    Runs the given coroutine concurrently. If the function is called from an event loop, you can await the result
    of the returned task. If not, then it blocks till the task is finished and returns the result of the coroutine.
    
    Parameters
    ----------
    coroutine : `GeneratorType`, `CoroutineType`
        The coroutine to create task from.
    
    loop : `None`, ``EventThread`` = `None`, Optional
        The event loop to schedule the created task on.
    
    Returns
    -------
    task : ``Task``, ``FutureAsyncWrapper``, `Any`
    
    Raises
    ------
    RuntimeError
        There are are no detectable event loops.
    BaseException
        Any exception raised by `coroutine`.
    """
    loop, is_current_thread = _get_event_loop_is_current_thread(loop)
    
    task = Task(coroutine, loop)
    
    if is_current_thread:
        return task
    
    thread = current_thread()
    if isinstance(thread, EventThread):
        # `.async_wrap` wakes up the event loop
        return task.async_wrap(thread)
    
    loop.wake_up()
    return task.sync_wrap().wait()



__all__ = ()

from ...utils import copy_docs

from ..protocols_and_transports import AbstractProtocolBase


class SubprocessWritePipeProtocol(AbstractProtocolBase):
    """
    Asynchronous subprocess write pipe protocol.
    
    Attributes
    ----------
    _disconnected : `bool`
        Whether the protocol is disconnected.
    _file_descriptor : `int`
        The used socket's file descriptor number.
    _transport : ``UnixWritePipeTransport``, `Any`
        Asynchronous transport implementation.
    _process : ``AsyncProcess``
        The parent process of the pipe protocol.
    """
    __slots__ = ('_disconnected', '_file_descriptor', '_transport', '_process', )
    
    def __new__(cls, process, file_descriptor):
        """
        Creates a new ``SubprocessWritePipeProtocol`` with the given parameters.
        
        Parameters
        ----------
        process : ``AsyncProcess``
            The parent process of the pipe protocol.
        file_descriptor : `int`
            The used socket's file descriptor number.
        """
        self = object.__new__(cls)
        self._process = process
        self._file_descriptor = file_descriptor
        self._transport = None
        self._disconnected = False
        return self
    
    
    @copy_docs(AbstractProtocolBase.connection_made)
    def connection_made(self, transport):
        self._transport = transport
    
    
    def __repr__(self):
        """Returns the subprocess write protocol's representation."""
        return f'<{self.__class__.__name__} file_descriptor={self._file_descriptor} pipe={self._transport!r}>'
    
    
    @copy_docs(AbstractProtocolBase.connection_lost)
    def connection_lost(self, exception):
        process = self._process
        if (process is not None):
            self._process = None
            process._pipe_connection_lost(self._file_descriptor, exception)
            self._disconnected = True
    
    
    @copy_docs(AbstractProtocolBase.pause_writing)
    def pause_writing(self):
        self._process.pause_writing()
    
    
    @copy_docs(AbstractProtocolBase.resume_writing)
    def resume_writing(self):
        self._process.resume_writing()


    @copy_docs(AbstractProtocolBase.get_transport)
    def get_transport(self):
        return self._transport


class SubprocessReadPipeProtocol(SubprocessWritePipeProtocol):
    """
    Asynchronous subprocess read pipe protocol.
    
    Attributes
    ----------
    _disconnected : `bool`
        Whether the protocol is disconnected.
    _file_descriptor : `int`
        The used socket's file descriptor number.
    _transport : ``UnixWritePipeTransport``, `Any`
        Asynchronous transport implementation.
    _process : ``AsyncProcess``
        The parent process of the pipe protocol.
    """
    __slots__ = ()
    
    @copy_docs(AbstractProtocolBase.data_received)
    def data_received(self, data):
        self._process._pipe_data_received(self._file_descriptor, data)
    
    
    @copy_docs(AbstractProtocolBase.eof_received)
    def eof_received(self):
        return False



__all__ = ('IgnoreCaseMultiValueDictionary', )

from .docs import copy_docs, has_docs
from .ignore_case_string import IgnoreCaseString
from .multi_value_dictionary import MultiValueDictionary


@has_docs
class IgnoreCaseMultiValueDictionary(MultiValueDictionary):
    """
    ``MultiValueDictionary`` subclass, what can be used to hold http headers.
    
    It's keys ignore casing.
    """
    __slots__ = ()
    
    @copy_docs(MultiValueDictionary.__init__)
    def __init__(self, iterable=None):
        dict.__init__(self)
        
        if (iterable is None) or (not iterable):
            return
        
        getitem = dict.__getitem__
        setitem = dict.__setitem__
        
        if type(iterable) is type(self):
            for key, values in dict.items(iterable):
                setitem(self, key, values.copy())
        
        elif isinstance(iterable, MultiValueDictionary):
            for key, values in dict.items(iterable):
                setitem(self, IgnoreCaseString(key), values.copy())
        
        elif isinstance(iterable, dict):
            for key, value in iterable.items():
                key = IgnoreCaseString(key)
                setitem(self, key, [value])
        
        else:
            for key, value in iterable:
                key = IgnoreCaseString(key)
                try:
                    values = getitem(self, key)
                except KeyError:
                    setitem(self, key, [value])
                else:
                    if value not in values:
                        values.append(value)
    
    
    @copy_docs(MultiValueDictionary.__getitem__)
    def __getitem__(self, key):
        key = IgnoreCaseString(key)
        return dict.__getitem__(self, key)[0]
    
    
    @copy_docs(MultiValueDictionary.__setitem__)
    def __setitem__(self, key, value):
        key = IgnoreCaseString(key)
        MultiValueDictionary.__setitem__(self, key, value)
    
    
    @copy_docs(MultiValueDictionary.__delitem__)
    def __delitem__(self, key):
        key = IgnoreCaseString(key)
        MultiValueDictionary.__delitem__(self, key)
    
    
    @copy_docs(MultiValueDictionary.extend)
    def extend(self, mapping):
        getitem = dict.__getitem__
        setitem = dict.__setitem__
        
        for key, value in mapping.items():
            key = IgnoreCaseString(key)
            try:
                values = getitem(self, key)
            except KeyError:
                setitem(self, key, [value])
            else:
                if value not in values:
                    values.append(value)
    
    
    @copy_docs(MultiValueDictionary.get_all)
    def get_all(self, key, default=None):
        key = IgnoreCaseString(key)
        return MultiValueDictionary.get_all(self, key, default)
    
    
    @copy_docs(MultiValueDictionary.get_one)
    def get_one(self, key, default=None):
        key = IgnoreCaseString(key)
        return MultiValueDictionary.get_one(self, key, default)
    
    get = get_one
    
    
    @copy_docs(MultiValueDictionary.setdefault)
    def setdefault(self, key, default=None):
        key = IgnoreCaseString(key)
        return MultiValueDictionary.setdefault(self, key, default)
    
    
    @copy_docs(MultiValueDictionary.pop_all)
    def pop_all(self, key, default=...):
        key = IgnoreCaseString(key)
        return MultiValueDictionary.pop_all(self, key, default)
    
    
    @copy_docs(MultiValueDictionary.pop_one)
    def pop_one(self, key, default=...):
        key = IgnoreCaseString(key)
        return MultiValueDictionary.pop_one(self, key, default)
    
    pop = pop_one



__all__ = ('DatagramAddressedReadProtocol', 'DatagramMergerReadProtocol', 'ReadProtocolBase', 'ReadWriteProtocolBase',)

from collections import deque

from ...utils import copy_docs

from ..exceptions import CancelledError
from ..traps import Future, Task, future_or_timeout, skip_ready_cycle

from .abstract import AbstractProtocolBase


CHUNK_LIMIT = 32

class ReadProtocolBase(AbstractProtocolBase):
    """
    Asynchronous read protocol implementation.
    
    Scarlet-io backend uses optimistic generator based chunked readers, which have really long and complicated
    implementation, but their speed is pretty good.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_waiter : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `Any`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    """
    __slots__ = (
        '_at_eof', '_chunks', '_exception', '_loop', '_offset', '_paused', '_payload_reader', '_payload_waiter',
        '_transport'
    )
    
    def __new__(cls, loop):
        """
        Creates a new read protocol instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop, what the protocol uses for it's asynchronous tasks.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._transport = None
        self._exception = None
        self._chunks = deque()
        self._offset = 0
        self._at_eof = False
        self._payload_reader = None
        self._payload_waiter = None
        self._paused = False
        return self
    
    
    def __repr__(self):
        """Returns the transport's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        if self._at_eof:
            repr_parts.append(' at eof')
            field_added = True
        else:
            field_added = False
        
        transport = self._transport
        if (transport is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' transport=')
            repr_parts.append(repr(transport))
        
        exception = self._exception
        if (exception is not None):
            if field_added:
                repr_parts.append(', ')
            else:
                field_added = True
            
            repr_parts.append(' exception=')
            repr_parts.append(repr(exception))
        
        payload_reader = self._payload_reader
        if (payload_reader is not None):
            if field_added:
                repr_parts.append(', ')
            
            repr_parts.append(' payload_reader=')
            repr_parts.append(repr(payload_reader))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(AbstractProtocolBase.connection_made)
    def connection_made(self, transport):
        self._transport = transport
    
    
    @copy_docs(AbstractProtocolBase.connection_lost)
    def connection_lost(self, exception):
        """
        Called when the connection is lost or closed.
        
        Parameters
        ----------
        exception : `None`, `BaseException`
            Defines whether the connection is closed, or an exception was received.
            
            If the connection was closed, then `exception` is given as `None`. This can happen at the case, when eof is
            received as well.
        """
        if exception is None:
            self.eof_received()
        else:
            self.set_exception(exception)
    
    
    @copy_docs(AbstractProtocolBase.get_extra_info)
    def get_extra_info(self, name, default=None):
        transport = self._transport
        if transport is None:
            info = default
        else:
            info = transport.get_extra_info(name, default)
        
        return info
    
    
    @copy_docs(AbstractProtocolBase.close)
    def close(self):
        transport = self._transport
        if (transport is not None):
            transport.close()
    
    
    @copy_docs(AbstractProtocolBase.close_transport)
    def get_transport(self):
        return self._transport
    
    
    # read related
    @property
    def size(self):
        """
        Returns the received and not yet processed data's size by the protocol.
        
        Returns
        -------
        size : `int`
        """
        result = - self._offset
        for chunk in self._chunks:
            result += len(chunk)
        
        return result
    
    
    def at_eof(self):
        """
        Returns whether the protocol is at eof. If at eof, but has content left, then returns `False` however.
        
        Returns
        -------
        at_eof : `bool`
        """
        if not self._at_eof:
            return False
        
        if (self._payload_reader is not None):
            return False
        
        if self.size:
            return False
    
        return True
    
    
    @copy_docs(AbstractProtocolBase.set_exception)
    def set_exception(self, exception):
        self._exception = exception
        
        payload_waiter = self._payload_waiter
        if (payload_waiter is None):
            return
        
        self._payload_waiter = None
        payload_waiter.set_exception_if_pending(exception)
        
        self._payload_reader.close()
        self._payload_reader = None
    
    
    @copy_docs(AbstractProtocolBase.eof_received)
    def eof_received(self):
        self._at_eof = True
        
        payload_reader = self._payload_reader
        if payload_reader is None:
            return False
        
        try:
             payload_reader.throw(CancelledError())
        except CancelledError as err:
            new_exception = ConnectionError('Connection closed unexpectedly with EOF.')
            new_exception.__cause__ = err
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            payload_waiter.set_exception_if_pending(new_exception)
        
        except StopIteration as err:
            args = err.args
            if not args:
                result = None
            elif len(args) == 1:
                result = args[0]
            else:
                result = args
            
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            payload_waiter.set_result_if_pending(result)
        except GeneratorExit as err:
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            exception = CancelledError()
            exception.__cause__ = err
            payload_waiter.set_exception_if_pending(exception)
        except BaseException as err:
            payload_waiter = self._payload_waiter
            self._payload_reader = None
            self._payload_waiter = None
            payload_waiter.set_exception_if_pending(err)
        
        return False
    
    
    @copy_docs(AbstractProtocolBase.data_received)
    def data_received(self, data):
        if not data:
            return
        
        payload_reader = self._payload_reader
        if (payload_reader is None):
            chunks = self._chunks
            chunks.append(data)
            if (len(chunks) > CHUNK_LIMIT) and (not self._paused):
                transport = self._transport
                if (transport is not None):
                    try:
                        transport.pause_reading()
                    except (AttributeError, NotImplementedError):
                        #cant be paused
                        self._transport = None
                    else:
                        self._paused = True
        else:
            try:
                payload_reader.send(data)
            except StopIteration as err:
                args = err.args
                if not args:
                    result = None
                elif len(args) == 1:
                    result = args[0]
                else:
                    result = args
                
                payload_waiter = self._payload_waiter
                self._payload_reader = None
                self._payload_waiter = None
                payload_waiter.set_result_if_pending(result)
            except GeneratorExit as err:
                payload_waiter = self._payload_waiter
                self._payload_reader = None
                self._payload_waiter = None
                exception = CancelledError()
                exception.__cause__ = err
                payload_waiter.set_exception_if_pending(exception)
            except BaseException as err:
                payload_waiter = self._payload_waiter
                self._payload_reader = None
                self._payload_waiter = None
                payload_waiter.set_exception_if_pending(err)
    
    
    def handle_payload_waiter_cancellation(self):
        """
        If you expect, that the payload waiter will be cancelled from outside, call this method to throw eof into the
        protocol at that case.
        """
        payload_waiter = self._payload_waiter
        if (payload_waiter is not None):
            payload_waiter.add_done_callback(self._payload_waiter_cancellation_callback)
    
    
    def _payload_waiter_cancellation_callback(self, future):
        """
        Callback to the ``.payload_waiter`` by ``.handle_payload_waiter_cancellation`` to throw eof into the
        ``.payload_reader`` task if payload waiter is cancelled from outside.
        
        Parameters
        ----------
        future : ``Future``
            The respective ``.payload_waiter``.
        """
        if future.is_cancelled():
            self.eof_received()
    
    
    def cancel_current_reader(self):
        """
        Cancels the current reader task of the protocol.
        
        Notes
        -----
        The used data by the payload reader task cannot be reused.
        """
        payload_reader = self._payload_reader
        if payload_reader is None:
            return
        
        self._payload_reader = None
        payload_reader.close()
        
        payload_waiter = self._payload_waiter
        self._payload_waiter = None
        payload_waiter.cancel()
    
    
    def set_payload_reader(self, payload_reader):
        """
        Sets payload reader to the protocol.
        
        Parameters
        ----------
        payload_reader : `GeneratorType`
            A generator, what gets control, every time a chunk is received, till it returns or raises.
        
        Returns
        -------
        payload_waiter : ``Future``
            Waiter, to what the result of the `payload_reader` is set.
        """
        assert self._payload_reader is None, 'Payload reader already set!'
        
        payload_waiter = Future(self._loop)
        exception = self._exception
        
        if (exception is None):
            try:
                payload_reader.send(None)
            except StopIteration as err:
                args = err.args
                if not args:
                    result = None
                elif len(args) == 1:
                    result = args[0]
                else:
                    result = args
                
                payload_waiter.set_result_if_pending(result)
            except GeneratorExit as err:
                exception = CancelledError()
                exception.__cause__ = err
                payload_waiter.set_exception_if_pending(exception)
            except BaseException as err:
                payload_waiter.set_exception_if_pending(err)
            
            else:
                self._payload_waiter = payload_waiter
                self._payload_reader = payload_reader
        else:
            payload_waiter.set_exception(exception)
        
        return payload_waiter
    
    
    async def read(self, n=-1):
        """
        Reads up to `n` amount of bytes from the protocol.
        
        If `n` is given as a negative integer, then will read until eof.
        
        This method is a coroutine.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Defaults to `-1`.
        
        Returns
        -------
        result : `bytes`
        """
        try:
            return await self.set_payload_reader(self._read_until_eof() if n < 0 else self._read_exactly(n))
        except EOFError as err:
            return err.args[0]
    
    
    async def read_exactly(self, n):
        """
        Read exactly `n` bytes from the protocol.
        
        This method is a coroutine.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read. Cannot be negative.
        
        Returns
        -------
        result : `str`
        
        Raises
        ------
        ValueError
            If `n` is less than `0`.
        EOFError
            Connection lost before `0` bytes were received.
        BaseException
            Connection lost exception if applicable.
        """
        exception = self._exception
        if (exception is not None):
            raise exception
        
        if n < 1:
            if n == 0:
                return b''
            
            raise ValueError(f'`.read_exactly` size can not be less than `0`, got {n!r}.')
        
        return await self.set_payload_reader(self._read_exactly(n))
    
    
    async def read_line(self):
        raise NotImplementedError
    
    
    async def read_until(self):
        raise NotImplementedError
    
    
    async def read_once(self):
        """
        Waits till exactly one chunk is of data is received.
        
        This method is a coroutine.
        
        Returns
        -------
        data : `bytes`
        
        Raises
        ------
        BaseException
            Connection lost exception if applicable.
        """
        exception = self._exception
        if (exception is not None):
            raise exception
        
        return await self.set_payload_reader(self._read_once())
    
    
    def _wait_for_data(self):
        """
        Payload reader task helper, what waits for 1 chunk to be receive, then adds it to ``._chunks`` and also returns
        it as well.
        
        This method is a generator.
        
        Returns
        -------
        chunk : `bytes`
        
        Raises
        ------
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        if self._paused:
            self._paused = False
            self._transport.resume_reading()
        
        chunk = yield
        self._chunks.append(chunk)
        return chunk
    
    
    def _read_exactly(self, n):
        """
        Payload reader task, what reads exactly `n` bytes from the protocol.
        
        This method is a generator.
        
        Parameters
        ----------
        n : `int`
            The amount of bytes to read.
        
        Returns
        -------
        collected : `bytes`
        
        Raises
        ------
        ValueError
            If `n` is given as a negative integer.
        EofError
            Connection lost before `n` bytes were received.
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        if n < 1:
            if n < 0:
                raise ValueError(f'`.read_exactly` called with negative `n`, got {n!r}.')
            else:
                return b''
        
        chunks = self._chunks
        if chunks:
            chunk = chunks[0]
            offset = self._offset
        else:
            if self._at_eof:
                raise EOFError(b'')
            
            chunk = yield from self._wait_for_data()
            offset = 0
        
        chunk_size = len(chunk)
        if offset == 0:
            if chunk_size > n:
                self._offset = n
                return chunk[:n]
            # chunk same size as the requested?
            elif chunk_size == n:
                del chunks[0]
                # offset is already 0, nice!
                return chunk
            
            else:
                n -= len(chunk)
                collected = [chunk]
                del chunks[0]
        else:
            end = offset + n
            if chunk_size > end:
                self._offset = end
                return chunk[offset:end]
            # chunk_size + offset end when the requested's end is.
            elif chunk_size == end:
                del chunks[0]
                self._offset = 0
                return chunk[offset:]
            
            else:
                n -= (chunk_size - offset)
                collected = [memoryview(chunk)[offset:]]
                del chunks[0]
        
        while True:
            if chunks:
                chunk = chunks[0]
            else:
                if self._at_eof:
                    self._offset = 0
                    raise EOFError(b''.join(collected))
                
                chunk = yield from self._wait_for_data()
            
            chunk_size = len(chunk)
            
            n -= chunk_size
            
            if n > 0:
                collected.append(chunk)
                del chunks[0]
                continue
            
            if n == 0:
                collected.append(chunk)
                del chunks[0]
                self._offset = 0
                return b''.join(collected)
            
            offset = self._offset = chunk_size + n
            collected.append(memoryview(chunk)[:offset])
            return b''.join(collected)
    
    
    def _read_until(self, boundary):
        """
        Payload reader task, what reads until `boundary` is hit.
        
        This method is a generator.
        
        Parameters
        ----------
        boundary : `bytes`
            The amount of bytes to read.
        
        Returns
        -------
        collected : `bytes`, `bytearray`
        
        Raises
        ------
        EofError
            Connection lost before `n` bytes were received.
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        # This method is mainly used for multipart reading, and we can forget optimizations usually.
        boundary_length = len(boundary)
        chunks = self._chunks
        if chunks:
            chunk = chunks[0]
            offset = self._offset
        else:
            if self._at_eof:
                raise EOFError(b'')
            
            chunk = yield from self._wait_for_data()
            offset = 0
        
        # Optimal case is when we instantly hit boundary.
        if len(chunk) > boundary_length:
            index = chunk.find(boundary, offset)
            if index != -1:
                # Barrier found
                data = chunk[offset:index]
                offset += boundary
                if offset == len(chunk):
                    del chunks[0]
                    offset = 0
                self._offset = offset
                return data
            
            offset = len(chunk) - boundary_length
        
        # Second case, we create a bytearray and push the data to it.
        data = bytearray(chunk)
        while True:
            if chunks:
                chunk = chunks[0]
            else:
                if self._at_eof:
                    raise EOFError(b'')
                
                chunk = yield from self._wait_for_data()
            
            data.extend(chunk)
            index = chunk.find(boundary, offset)
            if index != -1:
                # Barrier found
                offset = len(chunk) - len(data) + index + boundary_length
                if offset == len(chunk):
                    del chunks[0]
                    offset = 0
                
                self._offset = offset
                del data[index:]
                return data
            
            offset = len(data) - boundary_length
            del chunks[0]
    
    
    def _read_until_eof(self):
        """
        Payload reader task, which reads the protocol until EOF is hit.
        
        This method is a generator.
        
        Returns
        -------
        collected : `bytes`
        
        Raises
        ------
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        chunks = self._chunks
        
        if not self._at_eof:
            while True:
                try:
                    yield from self._wait_for_data()
                except (CancelledError, GeneratorExit):
                    if self._at_eof:
                        break
                    
                    raise
        
        if not chunks:
            return b''
        
        offset = self._offset
        if offset:
            chunks[0] = memoryview(chunks[0])[offset:]
        
        collected = b''.join(chunks)
        chunks.clear()
        return collected
    
    
    def _read_once(self):
        """
        Reader task for reading exactly one chunk out.
        
        This method is a generator.
        
        Returns
        -------
        chunk : `bytes`
            The read chunk. Returns empty `bytes` on eof.
        
        Raises
        ------
        CancelledError
            If the reader task is cancelled not by receiving eof.
        """
        chunks = self._chunks
        if not chunks:
            try:
                yield from self._wait_for_data()
            except (CancelledError, GeneratorExit):
                if self._at_eof:
                    return b''
                
                raise
        
        chunk = chunks.popleft()
        offset = self._offset
        if offset:
            self._offset = 0
            chunk = chunk[offset:]
        
        return chunk



class ReadWriteProtocolBase(ReadProtocolBase):
    """
    Read-write asynchronous protocol implementation.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_waiter : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `Any`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    _drain_waiter : `None`, ``Future``
        A future, what is used to block the writing task, till it's writen data is drained.
    """
    __slots__ = ('_drain_waiter', )
    
    def __new__(cls, loop):
        """
        Creates a new read-write protocol instance.
        
        Parameters
        ----------
        loop : ``EventThread``
            The respective event loop, what the protocol uses for it's asynchronous tasks.
        """
        self = ReadProtocolBase.__new__(cls, loop)
        
        self._drain_waiter = None
        
        return self
    
    
    @copy_docs(AbstractProtocolBase.pause_writing)
    def pause_writing(self):
        self._paused = True
    
    
    @copy_docs(AbstractProtocolBase.resume_writing)
    def resume_writing(self):
        self._paused = False
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        drain_waiter.set_result_if_pending(None)
    
    
    async def _drain_helper(self):
        """
        Called when the transport buffer after writing goes over the high limit mark to wait till it is drained.
        
        This method is a coroutine.
        """
        if not self._paused:
            return
        
        drain_waiter = Future(self._loop)
        self._drain_waiter = drain_waiter
        await drain_waiter
    
    @copy_docs(ReadProtocolBase.connection_lost)
    def connection_lost(self, exception):
        if exception is None:
            self.eof_received()
        else:
            self.set_exception(exception)
        
        # wake up the writer if currently paused.
        if not self._paused:
            return
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        if drain_waiter.is_done():
            return
        
        if exception is None:
            drain_waiter.set_result(None)
        else:
            drain_waiter.set_exception(exception)
    
    
    @copy_docs(AbstractProtocolBase.write)
    def write(self, data):
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self={self!r}.')
        
        transport.write(data)
    
    
    @copy_docs(AbstractProtocolBase.writelines)
    def writelines(self, lines):
        transport = self._transport
        if transport is None:
            raise RuntimeError(f'Protocol has no attached transport; self={self!r}.')
        
        transport.writelines(lines)
    
    
    @copy_docs(AbstractProtocolBase.write_eof)
    def write_eof(self):
        transport = self._transport
        if (transport is not None):
            transport.write_eof()
    
    
    @copy_docs(AbstractProtocolBase.can_write_eof)
    def can_write_eof(self):
        transport = self._transport
        if (transport is None):
            return False
        
        return transport.can_write_eof()
    
    
    @copy_docs(AbstractProtocolBase.drain)
    async def drain(self):
        # use after writing
        exception = self._exception
        if (exception is not None):
            raise exception
        
        transport = self._transport
        if (transport is not None):
            if transport.is_closing():
                # skip 1 loop, so connection_lost() will be called
                await skip_ready_cycle()
        
        await self._drain_helper()



class DatagramAddressedReadProtocol(AbstractProtocolBase):
    """
    Datagram reader protocol for reading from payloads form multiple addresses.
    
    Attributes
    ----------
    _by_address : `dict` of (`tuple` (`str`, `int`), ``ReadProtocolBase``) items
        Dictionary to store the alive readers by address.
    _loop : ``EventThread``
        The loop to what the protocol is bound to.
    _protocol_factory : `type`
        Protocol type to create for each address.
    _transport : `Any`
        Asynchronous transport implementation, what calls the protocol's ``.datagram_received`` when data is
        received.
    _waiters : `None`, `list` of `Future`
        Waiters for any payload receive if applicable.
    """
    __slots__ = ('_by_address', '_loop', '_protocol_factory', '_transport', '_waiters',)
    
    def __new__(cls, loop, protocol_factory):
        """
        Creates a new datagram addressed read protocol instance bound to the given loop.
        
        Parameters
        ----------
        loop : ``EventThread``
            The loop to what the protocol gonna be bound to.
        protocol_factory : `callable`
            Protocol type to create for each address.
            
            Defaults to ``ReadProtocolBase``.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._by_address = {}
        self._transport = None
        self._waiters = None
        self._protocol_factory = protocol_factory
        return self
    
    
    @copy_docs(AbstractProtocolBase.connection_made)
    def connection_made(self, transport):
        self._transport = transport
        for reader_protocol in self._by_address.values():
            reader_protocol.connection_made(transport)
    
    
    @copy_docs(AbstractProtocolBase.connection_lost)
    def connection_lost(self, exception):
        for reader_protocol in self._by_address.values():
            if exception is None:
                reader_protocol.eof_received()
            else:
                reader_protocol.set_exception(exception)
    
    
    @copy_docs(AbstractProtocolBase.datagram_received)
    def datagram_received(self, data, address):
        by_address = self._by_address
        try:
            protocol = by_address[address]
        except KeyError:
            protocol = self._protocol_factory()
            protocol.connection_made(self._transport)
            by_address[address] = protocol
        
        protocol.data_received(data)
        
        waiters = self._waiters
        if (waiters is not None):
            self._waiters = None
            result = (address, protocol)
            for waiter in waiters:
                waiter.set_result_if_pending(result)
    
    
    @copy_docs(AbstractProtocolBase.error_received)
    def error_received(self, exception):
        pass
    
    
    async def wait_for_receive(self, address=None, timeout=None):
        """
        Can be used to wait for payload to receive. Note, that this method should be used only initially, because the
        reader protocols implement the reading.
        
        This method is a coroutine.
        
        Parameters
        ----------
        address : `None`, `tuple` (`str`, `int`) = `None`, Optional
            The address of which payload is waiter for.
        timeout : `None`, `float = `None`, Optional
            The maximal amount of time to wait before raising `TimeoutError`.
        
        Returns
        -------
        result : `None`, ``ReadProtocolBase`` or (`tuple` (`str`, `int`), ``ReadProtocolBase``)
            - If `timeout` is given and timeout occur, then returns `None`.
            - if `address` is given and data is received from ir, then returns the respective ``ReadProtocolBase``.
            - If `address` is not given, then returns a `tuple` of the respective `address` and protocol.
        
        Raises
        ------
        TimeoutError
            - If timeout occured.
        """
        if timeout is None:
            if address is None:
                waiters = self._waiters
                if waiters is None:
                    self._waiters = waiters = []
                
                waiter = Future(self._loop)
                waiters.append(waiter)
                
                return await waiter
            else:
                while True:
                    waiters = self._waiters
                    if waiters is None:
                        self._waiters = waiters = []
                    
                    waiter = Future(self._loop)
                    waiters.append(waiter)
                    
                    address, protocol = await waiter
                    if address == address:
                        return protocol
                    
                    continue
        
        else:
            waiter = Task(self.wait_for_receive(address), self._loop)
            future_or_timeout(waiter, timeout)
            
            try:
                result = await waiter
            except TimeoutError:
                result = None
            
            return result
    
    
    @copy_docs(AbstractProtocolBase.close)
    def close(self):
        transport = self._transport
        if (transport is not None):
            transport.close()
    
    
    @copy_docs(AbstractProtocolBase.close_transport)
    def get_transport(self):
        return self._transport


class DatagramMergerReadProtocol(ReadProtocolBase):
    """
    Asynchronous read protocol implementation.
    
    Scarlet-io backend uses optimistic generator based chunked readers, which have really long and complicated
    implementation, but their speed is pretty good.
    
    Attributes
    ----------
    _at_eof : `bool`
        Whether the protocol received end of file.
    _chunks : `deque` of `bytes`
        Right feed, left pop queue, used to store the received data chunks.
    _exception : `None`, `BaseException`
        Exception set by ``.set_exception``, when an unexpected exception occur meanwhile reading from socket.
    _loop : ``EventThread``
        The event loop to what the protocol is bound to.
    _offset : `int`
        Byte offset, of the used up data of the most-left chunk.
    _paused : `bool`
        Whether the protocol's respective transport's reading is paused. Defaults to `False`.
        
        Also note, that not every transport supports pausing.
    _payload_reader : `None`, `GeneratorType`
        Payload reader generator, what gets the control back, when data, eof or any exception is received.
    _payload_waiter : `None` of ``Future``
        Payload waiter of the protocol, what's result is set, when the ``.payload_reader`` generator returns.
        
        If cancelled or marked by done or any other methods, the payload reader will not be cancelled.
    _transport : `None`, `Any`
        Asynchronous transport implementation. Is set meanwhile the protocol is alive.
    """
    __slots__ = ()
    
    @copy_docs(AbstractProtocolBase.datagram_received)
    def datagram_received(self, data, address):
        self.data_received(data)



__all__ = ()

import warnings

from ..core import EventThread, write_exception_maybe_async


HOOKS = []
LOADED_EXTENSIONS = set()

def register_library_extension(extension_name):
    """
    Registers a library extension, calling respective hooks if applicable.
    
    Parameters
    ----------
    extension_name : `str`
        The library extension's name.
    """
    if extension_name in LOADED_EXTENSIONS:
        warnings.warn(
            f'A library extension with name {extension_name!r} is already loaded.',
            stacklevel = 2,
        )
        return
    
    LOADED_EXTENSIONS.add(extension_name)
    
    for index in reversed(range(len(HOOKS))):
        requirements, hook = HOOKS[index]
        try:
            requirements.discard(extension_name)
        except KeyError:
            continue
        
        if requirements:
            continue
        
        del HOOKS[index]
        
        try:
            hook()
        except BaseException as err:
            write_exception_maybe_async(
                err,
                [
                    'register_library_extension(',
                    repr(extension_name),
                    ') ignores occurred exception meanwhile calling ',
                    repr(hook),
                    ' satisfied library extension hook.\n.'
                ]
            )


def add_library_extension_hook(hook, requirements):
    """
    Adds a library extension hook, what is called, when the given `requirements` are satisfied.
    
    Parameters
    ----------
    hook : `callable`
        Library extension hook called when all the required extensions are loaded as well.
    requirements : `iterable` of `str`
        An iterable of library extension names, which need to be resolved before calling the respective hook.
    """
    requirements_set = set(requirements)
    requirements_set.difference_update(LOADED_EXTENSIONS)
    
    if requirements_set:
        HOOKS.append((requirements_set, hook))
        return
    
    try:
        hook()
    except BaseException as err:
        write_exception_maybe_async(
            err,
            [
                'add_library_extension_hook(',
                repr(hook),
                ', ',
                repr(requirements),
                ') ignores occurred exception '
                'meanwhile calling ',
                repr(hook),
                ' satisfied library extension hook.\n.'
            ]
        )



__all__ = ()

from re import compile as re_compile


class PrefixTrimmer:
    """
    Prefix trimmer used to identify and remove prefixes.
    
    Attributes
    ----------
    excludes : `None`, `frozenset` of `str`
        Strings to exclude from matching.
    prefix_continuous_pattern : `re.Pattern`
        Prefix pattern matching continuous lines.
    prefix_initial_pattern : `re.Pattern`
        Prefix pattern matching only the initial line.
    prefix_last_pattern : `str`
        Prefix pattern matching the last line only if ``.prefix_continuous_pattern`` fails.
    """
    __slots__ = ('excludes', 'prefix_continuous_pattern', 'prefix_initial_pattern', 'prefix_last_pattern')
    
    def __new__(cls, prefix_initial_pattern, prefix_continuous_pattern, prefix_last_pattern, excludes ):
        """
        Creates a new prefix trimmer.
        
        Parameters
        ----------
        prefix_initial_pattern : `str`
            Prefix pattern matching only the initial line.
            
            > Not compiled.
        
        prefix_continuous_pattern : `str`
            Prefix pattern matching continuous lines.
            
            > Not compiled.
        
        prefix_last_pattern : `str`
            Prefix pattern matching the last line only if ``.prefix_continuous_pattern`` fails.
            
            > Not compiled.
        
        excludes  : `None`, `list` of `str`
            Strings to exclude from matching.
        """
        prefix_continuous_pattern = re_compile(prefix_continuous_pattern)
        prefix_initial_pattern = re_compile(prefix_initial_pattern)
        prefix_last_pattern = re_compile(prefix_last_pattern)
        
        if (excludes is not None):
            excludes = frozenset(excludes )
            if not excludes:
                excludes  = None
        
        self = object.__new__(cls)
        self.excludes  = excludes 
        self.prefix_continuous_pattern = prefix_continuous_pattern
        self.prefix_initial_pattern = prefix_initial_pattern
        self.prefix_last_pattern = prefix_last_pattern
        return self
    
    
    def checkout(self, lines):
        """
        Checks out the given lines for prefix.
        
        Parameters
        ----------
        lines : `list` of `str`
            The input lines to check out.
        
        Returns
        -------
        matched : `bool`
            Whether the given lines have the specified prefix.
        """
        line_count = len(lines)
        if line_count == 0:
            return False
        
        if line_count == 1:
            return self._checkout_one(lines)
        
        return self._checkout_more(lines)
    
    
    def _checkout_one(self, lines):
        """
        Checks out the given lines for prefix.
        
        This method is called by ``.checkout`` if `len(lines) == 1`
        
        Parameters
        ----------
        lines : `list` of `str`
            The input lines to check out.
        
        Returns
        -------
        matched : `bool`
            Whether the given lines have the specified prefix.
        """
        line = lines[0]
        
        for pattern in (self.prefix_initial_pattern, self.prefix_continuous_pattern):
            if pattern.match(line) is not None:
                return True
        
        if self.prefix_last_pattern.fullmatch(line) is not None:
            return True
        
        return False
    
    
    def _checkout_more(self, lines):
        """
        Checks out the given lines for prefix.
        
        This method is called by ``.checkout`` if `len(lines) > 1`
        
        Parameters
        ----------
        lines : `list` of `str`
            The input lines to check out.
        
        Returns
        -------
        matched : `bool`
            Whether the given lines have the specified prefix.
        """
        prefix_continuous_pattern = self.prefix_continuous_pattern
        
        index = 1
        length = len(lines)
        
        while index < length:
            line = lines[index]
            index += 1
            
            if prefix_continuous_pattern.match(line) is not None:
                continue
            
            if index < length:
                return False
            
            if self.prefix_last_pattern.fullmatch(line) is None:
                return False
            
            break
        
        return True
    
    
    def check_is_excluded(self, string):
        """
        Checks whether the given string is excluded from the trimmer.
        
        Parameters
        ----------
        string : `str`
            The string to check.
        
        Returns
        -------
        is_excluded : `bool`
        """
        excludes = self.excludes
        if (excludes is not None):
            if string in excludes:
                return True
        
        return False
    
    
    def apply(self, lines):
        """
        Applies the trimming to the given lines.
        
        Parameters
        ----------
        lines : `list` of `str`
            The lines to trim.
        
        Returns
        -------
        new_lines : `list` of `str`
            The new lines.
        """
        new_lines = [*self._iter_apply(lines)]
        
        for line in new_lines:
            if line:
                all_empty = False
                break
        else:
            all_empty = True
        
        if all_empty:
            new_lines.clear()
        
        return new_lines
    
    
    def _iter_apply(self, lines):
        """
        Applies trimming on the given lines.
        
        Called by ``.apply`` to generate the new lines.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        lines : `list` of `str`
            The lines to trim.
        
        Yields
        ------
        line : `str`
        """
        if lines:
            yield from self._iter_apply_line_first(lines)
            yield from self._iter_apply_line_rest(lines)
    
    
    def _iter_apply_line_first(self, lines):
        """
        Applies first line trimming on the first line of the given lines.
        
        Called by ``._iter_apply`` if `len(lines) > 0`.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        lines : `list` of `str`
            The lines to trim.
        
        Yields
        ------
        line : `str`
        """
        line = lines[0]
        
        for pattern in (self.prefix_initial_pattern, self.prefix_continuous_pattern):
            matched = pattern.match(line)
            if (matched is not None):
                line = line[matched.end():]
                break
        
        else:
            if (len(lines) == 1) and (self.prefix_last_pattern.fullmatch(line) is not None):
                return ''
        
        
        yield line
    
    
    def _iter_apply_line_rest(self, lines):
        """
        Applies continuous line trimming on all non-first lines.
        
        Called by ``._iter_apply`` if `len(lines) > 0`.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        lines : `list` of `str`
            The lines to trim.
        
        Yields
        ------
        line : `str`
        """
        prefix_continuous_pattern = self.prefix_continuous_pattern
        
        index = 1
        length = len(lines)
        
        while index < length:
            line = lines[index]
            index += 1
            
            matched = prefix_continuous_pattern.match(line)
            if (matched is not None):
                line = line[matched.end():]
            
            elif (index >= length) and (self.prefix_last_pattern.fullmatch(line) is not None):
                line = ''
            
            yield line


PREFIX_TRIMMERS = (
    PrefixTrimmer('In \[\d+\]\: ', ' +\.\.\.\: ', ' +\.\.\.\: ?', None), # scarletio
    PrefixTrimmer('\>\>\> ', '\.\.\. ', '\.\.\. ?', ['...']), # cpython
    PrefixTrimmer('\>\>\>\> ', '\.\.\.\. ', '\.\.\.\. ?', None), # pypy
)


def trim_console_prefix(content):
    """
    Parameters
    ----------
    content : `str`
        The source content.
    
    Returns
    -------
    trimmed_content : `None`, `str`
        Returns `None` if nothing was trimmed.
    """
    prefix_trimmers = [
        prefix_trimmer for prefix_trimmer in PREFIX_TRIMMERS if not prefix_trimmer.check_is_excluded(content)
    ]
    
    lines = content.splitlines()
    
    for prefix_trimmer in prefix_trimmers:
        if prefix_trimmer.checkout(lines):
            lines = prefix_trimmer.apply(lines)
            break
    
    else:
        return None
    
    return '\n'.join(lines)



__all__ = ('Event', )

from ...utils import copy_func, to_coroutine

from .future import Future


class Event:
    """
    Asynchronous equivalent to `threading.Event`.
    
    Attributes
    ----------
    _loop : ``EventThread``
        The event loop to what the event is bound.
    _value : `bool`
        The internal flag of the event, which defines, whether it is set.
    _waiters : `list` of ``Future``
        A list of futures waiting on the event to be set.
    """
    __slots__ = ('_loop', '_value', '_waiters',)
    
    def __new__(cls, loop):
        """
        Creates a new event object bound to the given event loop.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop to what the event is bound.
        """
        self = object.__new__(cls)
        self._loop = loop
        self._value = False
        self._waiters = []
        return self
    
    
    def is_set(self):
        """
        Returns whether the event's internal flag is set as `True`.
        
        Returns
        -------
        is_set: `bool`
        """
        return self._value
    
    
    def set(self):
        """
        Sets the event's internal flag to `True`, waking up all the tasks waiting for it.
        """
        if self._value:
            return
        
        self._value = True
        waiters = self._waiters
        for waiter in waiters:
            waiter.set_result_if_pending(None)
        
        waiters.clear()
    
    
    def clear(self):
        """
        Clears the internal flag of the event.
        """
        self._value = False
    
    
    def __iter__(self):
        """
        Waits util the event is set, or if it is already, returns immediately.
        
        This method is a generator. Should be used with `await` expression.
        """
        if self._value:
            return
        
        future = Future(self._loop)
        self._waiters.append(future)
        yield from future
    
    __await__ = __iter__
    
    wait = to_coroutine(copy_func(__iter__))
    
    def __repr__(self):
        """Returns the event's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' ',
        ]
        
        if self._value:
            state = 'set'
        else:
            state = 'unset'
        repr_parts.append(state)
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ('Handle', 'TimerHandle', 'TimerWeakHandle',)

from types import MethodType

from ...utils import WeakCallable, WeakReferer, ignore_frame, include, weak_method


ignore_frame(__spec__.origin, '_run', 'self.func(*self.args)', )
ignore_frame(__spec__.origin, 'run', 'handle._run()', )


write_exception_async = include('write_exception_async')


class Handle:
    """
    Object returned by a callback registration method:
    - ``EventThread.call_soon``
    - ``EventThread.call_soon_thread_safe``.
    
    Attributes
    ----------
    func : `callable`
        The wrapped function.
    args : `tuple` of `Any`
        Parameters to call ``.func`` with.
    cancelled : `bool`
        Whether the handle is cancelled.
    """
    __slots__ = ('func', 'args', 'cancelled',)
    
    def __init__(self, func, args):
        """
        Creates a new ``Handle`` with the given parameters.
        
        Parameters
        ----------
        func : `callable`
            The function. to wrap.
        args : `tuple` of `Any`
            Parameters to call `func` with.
        """
        self.func = func
        self.args = args
        self.cancelled = False
    
    
    def __repr__(self):
        """Returns the handle's representation."""
        return self._create_repr(False, None)
    
    
    def _get_repr_without(self, without):
        """
        Returns the representation of the handle without showing the representation of the given object.
        
        Parameters
        ----------
        without : `object`
            The value to ignore.
        
        Returns
        -------
        representation : `str`
        """
        return self._create_repr(True, without)
    
    
    def _create_repr(self, expect, expected_value):
        """
        Returns the representation of the handle.
        
        Parameters
        -----------
        expect : `bool`
            Whether `expected_value` should be ignored.
        expected_value : `object`
            The value to ignore.
        """

        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        if self.cancelled:
            repr_parts.append(' cancelled')
        
        else:
            repr_parts.append(' func=')
            func = self.func
            if expect and (func is expected_value):
                func_repr = '...'
            
            else:
                func_repr = repr(func)
            
            repr_parts.append(func_repr)
            repr_parts.append('(')
            
            parameters = self.args
            limit = len(parameters)
            if limit:
                index = 0
                while True:
                    parameter = parameters[index]
                    
                    repr_parts.append(' func=')
                    func = self.func
                    if expect and (func is expected_value):
                        parameter_repr = '...'
                    
                    else:
                        parameter_repr = repr(parameter)
                        
                    repr_parts.append(parameter_repr)
                    
                    index += 1
                    if index == limit:
                        break
                    
                    repr_parts.append(', ')
                    continue
            
            repr_parts.append(')')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def cancel(self):
        """Cancels the handle if not yet cancelled."""
        if not self.cancelled:
            self.cancelled = True
            self.func = None
            self.args = None
    
    
    def _run(self):
        """
        Calls the handle's function with it's parameters. If exception occurs meanwhile, renders it.
        
        Notes
        -----
        This method should be called only inside of an ``EventThread``.
        """
        try:
            self.func(*self.args)
        except BaseException as err:
            write_exception_async(
                err,
                [
                    'Exception occurred at ',
                    self.__class__.__name__,
                    '._run\nAt running ',
                    repr(self.func),
                    '\n',
                ]
            )
        
        self = None  # Needed to break cycles when an exception occurs.


class TimerHandle(Handle):
    """
    Object returned by a callback registration method:
    - ``EventThread.call_later``
    - ``EventThread.call_at``.
    
    Attributes
    ----------
    func : `callable`
        The wrapped function.
    args : `tuple` of `Any`
        Parameters to call ``.func`` with.
    cancelled : `bool`
        Whether the handle is cancelled.
    when : `float`
        The respective loop's time, when the handle should be called.
    """
    __slots__ = ('when',)
    
    def __init__(self, when, func, args):
        """
        Creates a new ``TimerHandle`` with the given parameters.
        
        Parameters
        ----------
        when : `float`
            The respective loop's time, when the handle should be called.
        func : `callable`
            The function. to wrap.
        args : `tuple` of `Any`
            Parameters to call `func` with.
        """
        self.func = func
        self.args = args
        self.cancelled = False
        self.when = when
    
    def __repr__(self):
        """Returns the timer handle's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        if self.cancelled:
            repr_parts.append(' cancelled')
        else:
            repr_parts.append(' func=')
            repr_parts.append(repr(self.func))
            repr_parts.append('(')
            
            args = self.args
            limit = len(args)
            if limit:
                index = 0
                while True:
                    arg = args[index]
                    repr_parts.append(repr(arg))
                    
                    index += 1
                    if index == limit:
                        break
                    
                    repr_parts.append(', ')
                    continue
            
            repr_parts.append(')')
            repr_parts.append(', when=')
            repr_parts.append(repr(self.when))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def __hash__(self):
        """Returns the hash of the time, when the handle will be called."""
        return hash(self.when)
    
    def __gt__(self, other):
        """Returns whether this timer handle should be called later than the other."""
        return self.when > other.when
    
    def __ge__(self, other):
        """Returns whether this timer handle should be called later than the other, or whether the two are equal."""
        if self.when > other.when:
            return True
        
        return self.__eq__(other)
    
    def __eq__(self, other):
        """Returns whether the two timer handles are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return (
            self.when       == other.when       and
            self.func       == other.func       and
            self.args       == other.args       and
            self.cancelled  == other.cancelled
        )
    
    def __ne__(self, other):
        """Returns whether the two timer handles are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return (
            self.when       != other.when       or
            self.func       != other.func       or
            self.args       != other.args       or
            self.cancelled  != other.cancelled
        )
    
    def __le__(self, other):
        """Returns whether this timer handle should be called earlier than the other, or whether the two are equal."""
        if self.when < other.when:
            return True
        
        return self.__eq__(other)
    
    def __lt__(self, other):
        """Returns whether this timer handle should be called earlier than the other."""
        return self.when < other.when


class TimerWeakHandle(TimerHandle):
    """
    Object returned by a callback registration method:
    - ``EventThread.call_later_weak``
    - ``EventThread.call_at_weak``.
    
    Used when the respective `func`, might be garbage collected before it callback would run.
    
    Attributes
    ----------
    func : `callable`
        The wrapped function.
    args : `tuple` of `Any`
        Parameters to call ``.func`` with.
    cancelled : `bool`
        Whether the handle is cancelled.
    when : `float`
        The respective loop's time, when the handle should be called.
    
    Notes
    -----
    This class also supports weakreferencing.
    """
    __slots__ = ('__weakref__', )
    
    def __init__(self, when, func, args):
        """
        Creates a new ``TimerWeakHandle`` with the given parameters.
        
        Parameters
        ----------
        when : `float`
            The respective loop's time, when the handle should be called.
        func : `callable`
            The function. to wrap.
        args : `tuple` of `Any`
            Parameters to call `func` with.
        
        Raises
        ------
        TypeError
            `func` is not weakreferable.
        """
        self.when = when
        callback = _TimerWeakHandleCallback(self)
        try:
            if type(func) is MethodType:
                func = weak_method.from_method(func, callback)
            else:
                func = WeakCallable(func, callback)
        except:
            # never leave a half finished object behind
            self.func = None
            self.args = None
            self.cancelled = True
            raise
        
        else:
            self.func = func
            self.args = args
            self.cancelled = False
    

class _TimerWeakHandleCallback:
    """
    Weakreference callback used by ``TimerWeakHandle`` to cancel the respective handle, when it's `func` gets
    garbage collected.
    
    Attributes
    ----------
    handle : ``WeakReferer`` (``TimerWeakHandle``)
        The respective handle.
    """
    __slots__ = ('handle', )
    
    def __init__(self, handle):
        """
        Creates a new weakreference callback used by ``TimerWeakHandle``.
        
        Parameters
        ----------
        handle : ``TimerWeakHandle``
            The respective handle.
        """
        self.handle = WeakReferer(handle)
    
    def __call__(self, reference):
        """
        Called, when the respective `TimerWeakHandle.func` is garbage collected.
        
        Parameters
        ----------
        reference : ``WeakReferer``
            Weakreference to the dead `func`.
        """
        handle = self.handle()
        if (handle is not None):
            handle.cancel()



__all__ = ('ClaimedExecutor', 'Executor', 'ExecutorThread', 'SyncQueue', 'SyncWait', )

import sys, warnings
from collections import deque
from threading import Event as SyncEvent, Lock as SyncLock, Thread, current_thread

from ...utils import ignore_frame, include
from ...utils.trace import render_exception_into

from ..exceptions import CancelledError
from ..time import LOOP_TIME
from ..traps import Future


EventThread = include('EventThread')
get_event_loop = include('get_event_loop')

ignore_frame(__spec__.origin, 'wait', 'raise exception',)
ignore_frame(__spec__.origin, 'get_result_no_wait', 'raise exception',)
ignore_frame(__spec__.origin, 'run', 'result = func()',)


EXECUTOR_RELEASE_INTERVAL = 0.6
EXECUTOR_RELEASE_MULTIPLIER = 2.5


class SyncWait:
    """
    Blocking waiter, what's result or exception can be set.
    
    Attributes
    ----------
    _result : `Any`
        The waiter's result if applicable. Defaults to `None`.
    _waiter : `threading.SyncEvent`
        Threading event, what is set, when the waiter's result is set.
    """
    __slots__ = ('_result', '_waiter',)
    
    def __init__(self):
        """
        Creates a new ``SyncWait``.
        """
        self._result = None
        self._waiter = SyncEvent()
    
    def set_result(self, result):
        """
        Sets the waiter's result.
        
        Parameters
        ----------
        result : `Any`
            The object to set as result.
        """
        self._result = result
        self._waiter.set()
    
    
    def wait(self):
        """
        Waits till the waiter's result is set, then returns it.
        
        Returns
        -------
        result : `Any`
        
        Raises
        ------
        TypeError
            The set exception is not a  `BaseException`.
        BaseException
            The set exception.
        """
        self._waiter.wait()
        return self._result


class SyncQueue:
    """
    Synchronous queue for retrieving set elements outside of an event loop.
    
    Attributes
    ----------
    _cancelled : `bool`
        Whether the queue is cancelled.
    _lock : `threading.SyncLock`
        Threading lock to grant exclusive access to the queue.
    _results : `deque`
        A deque of the results set to the queue.
    _waiter : `None`, ``SyncWait``
        Result waiter of the queue.
    """
    __slots__ = ('_cancelled', '_lock', '_results', '_waiter',)
    
    def __init__(self, iterable=None, max_length=None, cancelled=False):
        """
        Creates a new ``SyncQueue`` with the given parameters.
        
        Parameters
        ----------
        iterable : `None`, `iterable` of `Any` = `None`, Optional
            Iterable to set the queue's results initially from.
        max_length : `None`, `int` = `None`, Optional
            Maximal length of the queue. If the queue would pass it's maximal length, it's oldest results are popped.
        cancelled : `bool` = `False`, Optional
            Whether the queue should be initially cancelled.
        """
        self._results = deque(maxlen=max_length) if iterable is None else deque(iterable, maxlen=max_length)
        self._waiter = None
        self._cancelled = cancelled
        self._lock = SyncLock()
    
    
    def set_result(self, result):
        """
        Sets an element to the sync queue, what can be retrieved by ``.result`` or  by ``.result_no_wait``.
        
        Parameters
        ----------
        result : `Any`
            Sets a result to the sync queue.
        """
        with self._lock:
            # Should we raise InvalidStateError?
            waiter = self._waiter
            if waiter is None:
                self._results.append(result)
            else:
                waiter.set_result(result)
                self._waiter = None
    
    
    def cancel(self):
        """
        Cancels the sync-queue and it's waiter as well.
        """
        if self._cancelled:
            return
        
        with self._lock:
            # Should we raise InvalidStateError?
            self._cancelled = True
            
            waiter = self._waiter
            if waiter is not None:
                waiter.cancel()
                self._waiter = None
    
    
    def get_result(self):
        """
        Gets the next result of the sync queue. If the queue is empty, blocks till a new result is set with
        ``.set_result``.
        
        Returns
        -------
        result : `Any`
        """
        with self._lock:
            results = self._results
            if results:
                return results.popleft()
            
            waiter = self._waiter
            if waiter is None:
                if self._cancelled:
                    raise CancelledError()
                
                waiter = SyncWait()
                self._waiter = waiter
        
        return waiter.wait()
    
    
    @property
    def result(self):
        return self.get_result
    
    
    def get_result_no_wait(self):
        """
        Gets the next result of the sync queue instantly.
        
        Returns
        -------
        result : `Any`
        
        Raises
        ------
        CancelledError
            If the queue is empty and cancelled.
        IndexError
            If the queue is empty, but not cancelled.
        """
        with self._lock:
            results = self._results
            if results:
                return results.popleft()
            
            if self._cancelled:
                exception = CancelledError()
            else:
                raise IndexError('The queue is empty.')
            
            raise exception
    
    @property
    def result_no_wait(self):
        return self.get_result_no_wait
    
    
    def __repr__(self):
        """Returns the sync queue's representation."""
        with self._lock:
            repr_parts = [
                self.__class__.__name__,
                '([',
            ]
            
            results = self._results
            
            limit = len(results)
            if limit:
                index = 0
                while True:
                    result = results[index]
                    index += 1
                    
                    results.append(repr(result))
                    if index == limit:
                        break
                    
                    repr_parts.append(', ')
                    continue
            
            repr_parts.append(']')
            
            max_length = results.maxlen
            if (max_length is not None):
                repr_parts.append(', max_length=')
                repr_parts.append(repr(max_length))
            
            if self._cancelled:
                repr_parts.append(', cancelled')
            
            repr_parts.append(')')
            return ''.join(repr_parts)
    
    __str__ = __repr__
    
    # deque operations
    
    @property
    def max_length(self):
        """
        The maximal length of the queue if applicable.
        
        Returns
        -------
        max_length : `None`, `int`
        """
        return self._results.maxlen
    
    def clear(self):
        """
        Clears the queue.
        """
        self._results.clear()
    
    def copy(self):
        """
        Copies the queue.
        
        Returns
        -------
        new : ``SyncQueue``
        """
        with self._lock:
            new = object.__new__(type(self))
            
            new._results = self._results.copy()
            new._waiter = None
            new._cancelled = self._cancelled
            new._lock = SyncLock()
            
        return new
    
    def reverse(self):
        """
        Reverses the queue.
        """
        with self._lock:
            self._results.reverse()
    
    def __len__(self):
        """Returns the queue's length."""
        with self._lock:
            return len(self._results)
    
    def __bool__(self):
        """Returns `True` if the queue has any elements."""
        with self._lock:
            if self._results:
                result = True
            else:
                result = False
        
        return result

EXECUTOR_THREAD_CREATED = 0
EXECUTOR_THREAD_RUNNING = 1
EXECUTOR_THREAD_STOPPED = 2

class ExecutorThread(Thread):
    """
    A `threading.Thread` subclass for executing blocking code as an executor.
    
    Attributes
    ----------
    state : `int`
        Whether the executor thread is currently running.
        
        Can be on of the following:
        +---------------------------+-------+
        | Respective name           | Value |
        +===========================+=======+
        | EXECUTOR_THREAD_CREATED   | 0     |
        +---------------------------+-------+
        | EXECUTOR_THREAD_RUNNING   | 1     |
        +---------------------------+-------+
        | EXECUTOR_THREAD_STOPPED   | 2     |
        +---------------------------+-------+
    
    queue : ``SyncQueue`` of ``ExecutionPair``
        Synchronous queue of functions to execute and of their waiter future.
    """
    __slots__ = ('state', 'queue',)
    
    def __init__(self):
        """
        Creates and start a ``ExecutorThread``.
        """
        self.state = EXECUTOR_THREAD_CREATED
        self.queue = SyncQueue()
        Thread.__init__(self, daemon=True)
        self.start()
        
    def run(self):
        """
        The main runner of an ``ExecutorThread``.
        """
        if self.state != EXECUTOR_THREAD_CREATED:
            return
            
        self.state = EXECUTOR_THREAD_RUNNING
        
        queue = self.queue
        
        while self.state == EXECUTOR_THREAD_RUNNING:
            try:
                pair = queue.get_result()
                if pair is None:
                    return
                
                future = pair.future
                if future.is_done():
                    future = None
                    continue
                
                func = pair.func
                try:
                    result = func()
                except BaseException as err:
                    if isinstance(err, StopIteration):
                        exception = RuntimeError(f'{err.__class__.__name__} cannot be raised to a Future.')
                        exception.__cause__ = err
                        err = exception
                        exception = None
                    
                    future._loop.call_soon_thread_safe(future.__class__.set_exception_if_pending, future, err)
                else:
                    future._loop.call_soon_thread_safe(future.__class__.set_result_if_pending, future, result)
                    result = None
                
                future = None
                func = None
            
            except BaseException as err:
                extracted = [
                    self.__class__.__name__,
                    ' exception occurred\n',
                    repr(self),
                    '\n',
                ]
                render_exception_into(err, extend=extracted)
                sys.stderr.write(''.join(extracted))
    
    def execute(self, func, future=None):
        """
        Executes the given `func` on the executor thread.
        
        Parameters
        ----------
        func : `callable`
            The function to execute.
        future : `None`, ``Future`` = `None`, Optional
            A future, what's result is set, when the `func` finishes it's execution.
            
            If not given a new future is created. Defaults to `None`
        
        Returns
        -------
        future : ``Future``
            Result waiter future.
        
        Raises
        ------
        RuntimeError
            `future` parameter is not given an the method was not called from an event thread.
        """
        if future is None:
            loop = current_thread()
            if not isinstance(loop, EventThread):
                raise RuntimeError(
                    f'`future` parameter is not given and `{self!r}.execute` was not called '
                    f'from an `{EventThread.__name__}` either.'
                )
            
            future = Future(loop)
        
        self.queue.set_result(ExecutionPair(func, future,),)
        return future
    
    
    def cancel(self):
        """
        Cancels all the pending elements on the executor thread's queue, then cancels the queue itself as well.
        """
        self.state = EXECUTOR_THREAD_STOPPED
        queue = self.queue
        while queue:
            future = queue.get_result_no_wait().future
            future._loop.call_soon_thread_safe(future.cancel)
        
        queue.set_result(None)
    
    
    def release(self):
        """
        Cancels the executor thread's queue.
        """
        self.queue.set_result(None)
    
    
    def _shrink_queue(self):
        """
        Removes all the done elements from the executor's queue.
        """
        results = self.queue._results
        index = 0
        limit = len(results)
        while index < limit:
            future = results[index]
            if future.is_done():
                del results[index]
                limit -= 1
            else:
                index += 1


class ExecutionPair:
    """
    Stores a function to execute and it's future, to what the function's result or exception is set when done.
    
    Attributes
    ----------
    func : `callable`
        The function to execute.
    future : ``Future``
        The result waiter future to what ``.func``'s result or exception is set.
    """
    __slots__ = ('func', 'future', )
    
    
    def __init__(self, func, future):
        """
        Creates a new ``ExecutionPair``.
        
        Parameters
        ----------
        func : `callable`
            The function to execute.
        future : ``Future``
            The result waiter future to what ``.func``'s result or exception is set.
        """
        self.func = func
        self.future = future
    
    def __repr__(self):
        """Returns the execution pair's representation."""
        return f'{self.__class__.__name__}(func={self.func!r}, future={self.future!r})'


class _ClaimEndedCallback:
    """
    Future callback set to result waiter futures when calling ``ClaimedExecutor.release``.
    
    Gives back the executor thread, when it's queue becomes empty.
    
    Parameters
    ----------
    parent : ``Executor``
        The parent executor
    executor : ``ExecutorThread``
        The executor thread, what is given to the parent executor when all of it's tasks are done.
    """
    __slots__ = ('executor', 'parent',)
    
    def __init__(self, parent, executor):
        """
        Creates a new ``_ClaimEndedCallback`` with the given parameters.
        
        Parameters
        ----------
        parent : ``Executor``
            The parent executor
        executor : ``ExecutorThread``
            The executor thread, what is given to the parent executor when all of it's tasks are done.
        """
        self.parent = parent
        self.executor = executor
    
    def __call__(self, future):
        """Gives back the executor thread, when it's queue becomes empty."""
        loop = future._loop
        executor = self.executor
        executor._shrink_queue()
        if not executor.queue:
            parent = self.parent
            loop.call_soon_thread_safe(parent.__class__._claim_ended, parent, executor)
    
    def __eq__(self, other):
        """Returns whether the two claim ended callbacks are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.executor is other.executor


class ClaimedExecutor:
    """
    Wrapper for claimed executor threads.
    
    Attributes
    ----------
    executor : ``ExecutorThread``
        The claimed executor thread.
    parent : ``Executor``
        The parent executor from where the ``.executor`` is claimed.
    """
    __slots__ = ('executor', 'parent',)
    
    def __init__(self, parent, executor):
        """
        Creates a new claimed executor instance which will wrap the given executor thread of it's executor.
        
        Parameters
        ----------
        parent : ``Executor``
            The parent executor from where the ``.executor`` is claimed.
        executor : ``ExecutorThread``
            The claimed executor thread.
        """
        self.parent = parent
        self.executor = executor
    
    def execute(self, func):
        """
        Runs the given function in an executor thread. When the function is done, it's result or exception is set to
        the returned future.
        
        Parameters
        ----------
        func : ``Callable``
            The function to run inside of an executor thread.
        
        Returns
        -------
        future : ``Future``
            The future, to what the returned value or the raised exception of `func` is set.
        
        Raises
        ------
        RuntimeError
            The claimed executor is closed.
        """
        executor = self.executor
        if executor is None:
            raise RuntimeError(
                f'Executing on an already closed `{self.__class__.__name__}`.'
            )
        
        future = self.parent.create_future()
        
        executor.queue.set_result(ExecutionPair(func, future,),)
        
        return future
    
    def release(self):
        """
        Releases the claimed executor. Gives back it's executor thread to the parent executor, when all of it's
        tasks finish.
        """
        executor = self.executor
        if executor is None:
            return
        
        executor._shrink_queue()
        
        results = executor.queue._results
        if results:
            callback = _ClaimEndedCallback(self.parent, executor)
            for element in results:
                future = element.future
                if future.is_done():
                    continue
                
                future.add_done_callback(callback)
        else:
            self.parent._claim_ended(executor)
        
        self.executor = None
    
    __del__ = release
    
    def __enter__(self):
        """Entering a claimed executor returns itself."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        """Releases the claimed executor."""
        self.release()
        return False


class ExecutionEndedCallback:
    """
    Future callback set to result waiter futures when calling ``Executor.run_in_executor``.
    
    Parameters
    ----------
    parent : ``Executor``
        The parent executor
    executor : ``ExecutorThread``
        The executor thread, what is given to the parent executor when it's current task is done.
    """
    __slots__ = ('executor', 'parent',)
    
    def __init__(self, parent, executor):
        """
        Creates a new ``ExecutionEndedCallback`` with the given parameters.
        
        Parameters
        ----------
        parent : ``Executor``
            The parent executor
        executor : ``ExecutorThread``
            The executor thread, what is given to the parent executor when it's current task is done.
        """
        self.parent = parent
        self.executor = executor
        
    def __call__(self, future):
        """calls the parent executor's ``._execution_ended`` method, giving the executor thread back to it."""
        parent = self.parent
        future._loop.call_soon_thread_safe(parent.__class__._execution_ended, parent, self.executor)
    
    def __eq__(self, other):
        """Returns whether the two execution ended callbacks are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.executor is other.executor


class Executor:
    """
    Executor enabling running blocking sync code from async.
    
    Attributes
    ----------
    _kept_executor_count : `int`
        The minimal amount of executors to keep alive (or not close).
    _kept_executor_last_schedule : `float`
        When was last time cleanup scheduled.
    _kept_executor_release_handle : `None`, ``TimerHandle``
        Executor release handle.
    claimed_executors : `set` of ``ExecutorThread``
        Claimed executors, which are given back to the executor on release.
    free_executors : `deque`
        The free (or not used) executors of the executor.
    running_executors : `set` of ``ExecutorThread``
        The executors under use.
    """
    __slots__ = (
        '_kept_executor_count', '_kept_executor_last_schedule', '_kept_executor_release_handle', 'claimed_executors',
        'free_executors', 'running_executors'
    )
    
    def __init__(self, keep_executor_count=...):
        """
        Initializes the executor.
        
        Parameters
        ----------
        keep_executor_count : `int` = `1`, Optional
            The minimal amount of executors to keep alive (or not close).
        """
        if keep_executor_count is not ...:
            warnings.warn(
                (
                    f'`{self.__class__.__name__}`\'s `keep_executor_count` parameter is deprecated and will be removed '
                    f'in 2023 January. The kept executor count is now auto-calculated.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
        
        self._kept_executor_count = 0
        self._kept_executor_last_schedule = 0.0
        self._kept_executor_release_handle = None
        self.claimed_executors = set()
        self.free_executors = deque()
        self.running_executors = set()
        
    
    
    def __repr__(self):
        """Returns the executor's representation."""
        return f'<{self.__class__.__name__} free={self.get_free_executor_count()}, used={self.get_used_executor_count()}>'
    
    
    @property
    def used_executor_count(self):
        """
        Returns how much executor threads of the executors are currently under use.
        
        Returns
        -------
        executor_count : `int`
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.used_executor_count` parameter is deprecated and will be removed in '
                f'2023 January. Please use `.get_used_executor_count()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.get_used_executor_count()
    
    
    def get_used_executor_count(self):
        """
        Returns how much executor threads of the executors are currently under use.
        
        Returns
        -------
        executor_count : `int`
        """
        return len(self.running_executors) + len(self.claimed_executors)
    
    
    @property
    def free_executor_count(self):
        """
        Returns how much executor threads of the executors are currently under use.
        
        Returns
        -------
        executor_count : `int`
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.free_executor_count` parameter is deprecated and will be removed in '
                f'2023 January. Please use `.get_free_executor_count()` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.get_free_executor_count()
    
    
    def get_free_executor_count(self):
        """
        Returns how much free executors the executor has.
        
        Returns
        -------
        executor_count : `int`
        """
        return len(self.free_executors)
    
    
    def get_total_executor_count(self):
        """
        Returns the total amount of executors.
        
        Returns
        -------
        executor_count : `int`
        """
        return len(self.running_executors) + len(self.claimed_executors) + len(self.free_executors)
        
    
    def cancel_executors(self):
        """
        Cancels the executor threads of the executor.
        
        Raises ``CancelledError`` to the not yet started tasks.
        """
        self._reset_kept_executor_count()
        
        executors = self.free_executors
        while executors:
            executor = executors.pop()
            executor.release()
        
        executors = self.running_executors
        while executors:
            executor = executors.pop()
            executor.cancel()
        
        executors = self.claimed_executors
        while executors:
            executor = executors.pop()
            executor.cancel()
    
    
    def release_executors(self):
        """
        Releases the executor threads of the executor.
        
        Raises no exception to the already started tasks,
        """
        self._reset_kept_executor_count()
        
        executors = self.free_executors
        while executors:
            executor = executors.pop()
            executor.release()
        
        executors = self.running_executors
        while executors:
            executor = executors.pop()
            executor.release()
        
        executors = self.claimed_executors
        while executors:
            executor = executors.pop()
            executor.release()
    
    
    __del__ = release_executors
    
    
    def _reset_kept_executor_count(self):
        """
        Resets the kept executor count of the executor.        
        """
        self._kept_executor_count = 0
        
        kept_executor_release_handle = self._kept_executor_release_handle
        if (kept_executor_release_handle is not None):
            self._kept_executor_release_handle = None
            kept_executor_release_handle.cancel()
    
    
    def create_future(self):
        """
        Creates a future bound to the respective event loop.
        
        Returns
        -------
        future : ``Future``
            The created future.
        
        Raises
        ------
        RuntimeError
            If the method was not called from an ``EventLoop``.
        """
        local_thread = current_thread()
        if not isinstance(local_thread, EventThread):
            raise RuntimeError(
                f'`{self!r}.create_future` was not called from an `{EventThread.__name__}`, but from '
                f'{local_thread.__class__.__name__}; {local_thread!r}.'
            )
        
        return Future(local_thread)
    
    
    def run_in_executor(self, func):
        """
        Runs the given function in an executor thread. When the function is done, it's result or exception is set to
        the returned future.
        
        Parameters
        ----------
        func : ``Callable``
            The function to run inside of an executor thread.
        
        Returns
        -------
        future : ``Future``
            The future, to what the returned value or the raised exception of `func` is set.
            
            If the executor is cancelled, then these futures are cancelled as well.
        """
        future = self.create_future()
        executor = self._get_free_executor()
        self.running_executors.add(executor)
        future.add_done_callback(ExecutionEndedCallback(self, executor))
        executor.queue.set_result(ExecutionPair(func, future,),)
        return future
    
    
    def _execution_ended(self, executor):
        """
        Called when an executed function finishes ensured by ``.run_in_executor``. Used to give back the
        ``ExecutorThread`` to the executor and maybe cancel it.
        
        Parameters
        ----------
        executor : ``ExecutorThread``
            Executor with finished execution.
        """
        try:
            self.running_executors.remove(executor)
        except KeyError:
            return
        
        self._sync_keep(executor)
    
    
    def claim_executor(self):
        """
        Claims an executor thread from the executor. The thread is given back to the executor, when the returned
        wrapper is cancelled or released.
        
        Returns
        -------
        claimed_executor : ``ClaimedExecutor``
        """
        executor = self._get_free_executor()
        self.claimed_executors.add(executor)
        return ClaimedExecutor(self, executor)
    
    
    def _claim_ended(self, executor):
        """
        Called when a claimed executor is released and all of it's tasks are done. Used to give back the
        ``ExecutorThread`` to the executor and maybe cancel it.
        
        Parameters
        ----------
        executor : ``ExecutorThread``
            The given back executor.
        """
        try:
            self.claimed_executors.remove(executor)
        except KeyError:
            return
        
        self._sync_keep(executor)
    
    
    def _get_free_executor(self):
        """
        Gets a free executor thread from the executor. If there are no free executor threads, starts a new one.
        
        Returns
        -------
        executor : ``ExecutorThread``
        """
        free_executors = self.free_executors
        if free_executors:
            executor = free_executors.pop()
        else:
            executor = ExecutorThread()
        
        return executor
    
    
    def call_at(self, *args):
        """
        Placeholder method.
        
        Forwards all the parameters to the local ``EventThread``'s `.call_at` method.
        
        Parameters
        ----------
        *args : parameters
            The forwarded parameters.
        
        Returns
        -------
        handle : `None`, ``TimerHandle``
        """
        return get_event_loop().call_at(*args)
    
    
    def _sync_keep(self, executor):
        """
        Called when an executor thread is given back to the executor.
        
        Parameters
        ----------
        executor : ``ExecutorThread``
            The given back executor.
        """
        self.free_executors.append(executor)
        
        previously_used_executor_count = len(self.running_executors) + 1
        if previously_used_executor_count < self._kept_executor_count:
            return
        
        self._kept_executor_count = previously_used_executor_count
        current_time = LOOP_TIME()
        self._kept_executor_last_schedule = current_time
        
        handle = self._kept_executor_release_handle
        if (handle is None):
            self._kept_executor_release_handle = self.call_at(
                current_time + EXECUTOR_RELEASE_INTERVAL,
                type(self)._release_executor_step,
                self,
                current_time,
                EXECUTOR_RELEASE_INTERVAL,
            )
    
    
    def _release_executor_step(self, schedule_time, schedule_interval):
        """
        The method decides whether the executor should be kept or released.
        
        Parameters
        ----------
        schedule_time : `bool`
            When the step was scheduled.
        schedule_interval : `float`
            The interval between the new and the last scheduling.
        """
        last_schedule = self._kept_executor_last_schedule
        if last_schedule <= schedule_time:
            free_executors = self.free_executors
            if free_executors:
                executor = free_executors.pop()
                executor.release()
                
                kept_executor_count = self._kept_executor_count
                if kept_executor_count <= 0:
                    self._kept_executor_release_handle = None
                    return
                    
                self._kept_executor_count = kept_executor_count - 1
                current_time = LOOP_TIME()
                self._kept_executor_last_schedule = current_time
                
                self._kept_executor_release_handle = self.call_at(
                    current_time + EXECUTOR_RELEASE_INTERVAL,
                    type(self)._release_executor_step,
                    self,
                    current_time,
                    EXECUTOR_RELEASE_INTERVAL,
                )
            return
            
            # should not happen
            last_schedule = LOOP_TIME()
            self._kept_executor_last_schedule = last_schedule
        
        else:
            schedule_interval *= EXECUTOR_RELEASE_MULTIPLIER
        
        # Reschedule on the same way
        self._kept_executor_release_handle = self.call_at(
            last_schedule + schedule_interval,
            type(self)._release_executor_step,
            self,
            last_schedule,
            schedule_interval,
        )
